#!/usr/bin/env python
#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001, 2002
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: runlyntin.py.pyw,v 1.1 2003/10/03 02:17:48 willhelm Exp $
#######################################################################
bootoptions = {"ui": "tk",
               "datadir": "",
               "moduledir": [],
               "readfile": [],
               "snoopdefault": 1}

if __name__ == '__main__':
  import lyntin.engine
  lyntin.engine.main(bootoptions)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: action.py,v 1.24 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines the ActionManager which handles managing actions 
(triggers), matching triggers in mud_data and executing the resulting
action on behalf of the user for that session.

The ActionManager contains an ActionData object for every session that
has actions.

An action consists of:

1. the trigger statement
2. the response statement
3. the priority of the action
4. whether or not the action is a onetime action
5. optional action tag - the name of the action or action group

We also store a compiled regular expression of the trigger which
we use on incoming mud_data to check for triggered actions.

The compiled regular expressions gets recompiled every time a variable
changes--this allows us to handle Lyntin variables in the action trigger
statements.
"""
import re
from lyntin import manager, utils, event, exported, ansi
from lyntin.modules import modutils


# the placement variable regular expression
VARREGEXP = re.compile('%_?(\d+)')

class ActionData:
  def __init__(self, ses):
    self._actions = {}
    self._ses = ses
    self._disabled = {}
    self._actionlist = None

  def addAction(self, trigger, response, color=0, priority=5, onetime=0, tag=None):
    """
    Compiles a trigger pattern and adds the entire action to the
    hash.

    @param trigger: the trigger pattern
    @type  trigger: string

    @param response: what to do when the trigger pattern is found
    @type  response: string

    @param color: whether (1) or not (0) we try matching the line with
        the ansi colors still in it.  (i.e. if color==0, we filter the
        ansi colors out before matching)
    @type  color: boolean

    @param priority: the priority to run this action at.  actions
        are sorted by priority then by the trigger statement when
        we go to check for triggered actions.  default is 5
    @type  priority: int

    @param onetime: if the trigger is found, should this action then
        get removed after the response is executed
    @type  onetime: boolean

    @return: 1
    @rtype:  boolean
    """
    expansion = exported.expand_ses_vars(trigger, self._ses)
    if not expansion:
      expansion = trigger
    compiled = utils.compile_regexp(expansion, 1)
    self._actions[trigger] = (trigger, compiled, response, color, priority, onetime, tag)
    self._actionlist = None       # invalidating action list
    return 1

  def _recompileRegexps(self):
    """
    When a variable changes, we go through and recompile all the
    regular expressions for the actions in this session.
    """
    for mem in list(self._actions.keys()):
      (trigger, compiled, response, color, priority, onetime, tag) = self._actions[mem]
      expansion = exported.expand_ses_vars(trigger, self._ses)
      if not expansion:
        expansion = trigger

      compiled = utils.compile_regexp(expansion, 1)

      self._actions[trigger] = (trigger, compiled, response, color, priority, onetime, tag)

  def clear(self):
    """
    Clears all the stored actions from the action manager.
    """
    self._actions.clear()
    self._disabled = {}
    self._actionlist = None

  def getInfoMappings(self):
    l = []
    for key in list(self._actions.keys()):
      mem = self._actions[key]
      l.append( { "trigger": mem[0],
                  "action": mem[2],
                  "tag": mem[6],
                  "color": mem[3],
                  "priority": mem[4],
                  "onetime": mem[5] } )
    return l

  def removeActions(self, text, mytag=None):
    """
    Removes actions that match the given text and have given tag 
    from the list and returns the list of actions that were removed
    so the calling function knows what actually happened.

    @param text: all actions that match this text pattern will 
        be removed.  the text pattern is "expanded" by 
        "utils.expand_text"
    @type  text: string

    @param mytag: all actions with given tag will be removed.
    @type  mytag: string

    @return: list of tuples (trigger, response, tag) of the action
        that were removed.
    @rtype: (string, string, string)
    """
    actions = self._actions
    keys = []
    if text:
      keys = utils.expand_text(text, list(actions.keys()))
    elif mytag:  
      keys = list(actions.keys())

    ret = []
    for mem in keys:
      (trigger, compiled, response, color, priority, onetime, tag) = actions[mem]
      if not mytag or mytag == tag:
        ret.append((trigger, response, tag))
        del actions[mem]

    self._actionlist = None       # invalidating action list

    return ret

  def checkActions(self, text):
    """
    Checks to see if text triggered any actions.  Any resulting 
    actions will get added as an InputEvent to the queue.

    @param text: the data coming from the mud to check for triggers
    @type  text: string
    """
    # FIXME - make sure this works even when lines are broken up.

    actionlist = self._actionlist
    if not actionlist:
      #actionlist = list(filter(lambda x: x[6] not in self._disabled, self._actions.values()))
      #refactored for python 3
      actionlist = list([x for x in list(self._actions.values()) if x[6] not in self._disabled])

      #changing the following
      #actionlist.sort(lambda x,y:cmp(x[3], y[3]))
      actionlist.sort(key=lambda x: x[3])

      self._actionlist = actionlist

    colorline = utils.filter_cm(text)
    nocolorline = ansi.filter_ansi(colorline)

    # go through all the lines in the data and see if we have
    # any matches
    for (action, actioncompiled, response, color, priority, onetime, tag) in actionlist:
      if color:
        match = actioncompiled.search(colorline)
        line = colorline
      else:
        match = actioncompiled.search(nocolorline)
        line = nocolorline

      if match:
        # for every match we figure out what the expanded response
        # is and add it as an InputEvent in the queue.  the reason
        # we do a series of separate events rather than one big
        # event with ; separators is due to possible issues with 
        # braces and such in malformed responses.

        # get variables from the action
        actionvars = get_ordered_vars(action)

        # fill in values for all the variables in the match
        varvals = {}
        for i in range(len(actionvars)):
          varvals[actionvars[i]] = match.group(i+1)

        # add special variables
        varvals['a'] = line.replace(';', '_')
            
        # fill in response variables from those that
        # matched on the trigger
        response = utils.expand_vars(response, varvals)

        # event.InputEvent(response, internal=1, ses=self._ses).enqueue()
        try:
          exported.lyntin_command(response, internal=1, session=self._ses)
        except:
          exported.write_traceback()

        if onetime and action in self._actions:
          del self._actions[action]
          self._actionlist = None           # invalidate the list


  def getStatus(self):
    """
    Returns a one-liner as to how many actions we have.

    @return: a description of the status of this manager
    @rtype:  string
    """
    return "%d action(s)." % len(self._actions)

  def getInfo(self, text="", tag=None):
    """
    Returns information about the actions in here.

    This is used by #action to tell all the actions involved
    as well as #write which takes this information and dumps
    it to the file.

    @param text: the text to expand to find actions the user
        wants information about.
    @type  text: string

    @param tag: the tag which to find actions for.
    @type  tag: string

    @return: a list of strings where each string represents an action
    @rtype: list of strings
    """
    listing = list(self._actions.keys())
    if text:
      listing = utils.expand_text(text, listing)

    data = []
    for mem in listing:
      actup = self._actions[mem]
      
      if not tag or actup[6] == tag:
        data.append("action {%s} {%s} color={%d} priority={%d} onetime={%s} tag={%s}" % 
                (utils.escape(mem), utils.escape(actup[2]), actup[3], actup[4], actup[5], actup[6]))

    return data

  def getDisabledInfo(self, tag=None):
    """
    Returns information about disabled tags.

    @param tag: the tag which to find actions for.
    @type  tag: string

    @return: a list of strings where each string represents an action
    @rtype: list of strings
    """
    data = []
    for mem in list(self._disabled.keys()):
      if not tag or mem == tag:
        data.append("disable tag={%s}" % mem)

    return data

  def enable(self, tag):
    """
    Enables all the actions with given tag.

    @param tag: tag name
    @type tag: string
    """
    if tag in self._disabled:
      del self._disabled[tag]
      self._actionlist = None

  def disable(self, tag):
    """
    Disables all the actions with given tag.

    @param tag: tag name
    @type tag: string
    """
    self._disabled[tag] = 1
    self._actionlist = None

  def listTags(self):
    """
    Lists all the existing tags
    """
    tags = {}
    for action in list(self._actions.values()):
      tags[action[6]] = 0
    tags.update(self._disabled)  
    return [ "%s tag={%s}" % ((" enabled", "disabled")[disabled], mem)
             for (mem, disabled) in list(tags.items()) ]


class ActionManager(manager.Manager):
  def __init__(self):
    self._actions = {}

  def getActionData(self, ses):
    if ses not in self._actions:
      self._actions[ses] = ActionData(ses)
    return self._actions[ses]

  def clear(self, ses):
    if ses in self._actions:
      self._actions[ses].clear()

  def getInfoMappings(self, item, ses):
    if item != "action":
      raise ValueError("%s is not a valid item for this manager." % item)

    return self.getActionData(ses).getInfoMappings()

  def getItems(self):
    return [ "action" ]

  def getParameters(self, item):
    if item != "action":
      raise ValueError("%s is not a valid item for this manager." % item)

    return [ ("trigger", "Text that triggers the action."),
             ("action", "Command to execute when the trigger is kicked off."),
             ("tag", "Group of actions this action belongs to."),
             ("color", "Whether we try to match the line with color or not."),
             ("priority", "The priority to test this trigger at."),
             ("onetime", "Whether this action should be removed after it's triggered.")]
    
  def getInfo(self, ses, text="", tag=None):
    return self.getActionData(ses).getInfo(text, tag)

  def getDisabledInfo(self, ses, tag=None):
    return self.getActionData(ses).getDisabledInfo(tag)

  def listTags(self, ses):
    return self.getActionData(ses).listTags()

  def addSession(self, newsession, basesession=None):
    if basesession:
      if basesession in self._actions:
        bdata = self.getActionData(basesession)
        ndata = self.getActionData(newsession)

        for (mem, act) in list(bdata._actions.items()):
          ndata.addAction(mem, *act[2:])
        for tag in list(bdata._disabled.keys()):
          ndata.disable(tag)

  def removeSession(self, ses):
    if ses in self._actions:
      del self._actions[ses]

  def getStatus(self, ses):
    return self.getActionData(ses).getStatus()

  def persist(self, args):
    """
    write_hook function for persisting the state of our session.
    """
    ses = args["session"]
    quiet = args["quiet"]

    ad = self.getActionData(ses)

    data = ad.getInfo() + ad.getDisabledInfo()

    if quiet == 1:
      data = [m + " quiet={true}" for m in data]

    return data

  def variableChange(self, args):
    """
    When a variable changes, we need to recompile the regular
    expressions involved.  This facilitates that.

    This is registered with the variable_change hook.
    """
    ses = args["session"]
    if ses in self._actions:
      self._actions[ses]._recompileRegexps()

  def mudfilter(self, args):
    """
    mud_filter_hook function to check for actions when data
    comes from the mud.
    """
    ses = args["session"]
    text = args["dataadj"]

    if exported.get_config("ignoreactions", ses, 0) == 0:
      if ses in self._actions:
        self._actions[ses].checkActions(text)

    return text


def get_ordered_vars(text):
  """
  Takes in a string and removes any ordered variables
  from it.  Returns a list of the variables.

  @param text: the incoming string which may have ordered variables in it.
  @type  text: string

  @return: list of strings of the form '%[0-9]+' for ordered variable
      substitution.
  @rtype: list of strings
  """
  keylist = []
  matches = VARREGEXP.findall(text)

  for match in matches:
    keylist.append(match)

  return keylist

commands_dict = {}

def action_cmd(ses, args, input):
  """
  With no trigger, no action and no tag, prints all actions.
  With no trigger and no action, prints all actions with given tag.
  With a trigger and no action, prints actions that match the
  trigger statement.
  With a trigger and an action, creates an action.

  When data from the mud matches the trigger clause, the response
  will be executed.  Trigger clauses can use anchors (^ and $)
  to anchor the text to the beginning and end of the line 
  respectively.

  Triggers can also contain Lyntin pattern-variables which start
  with a % sign and have digits: %0, %1, %10...  When Lyntin sees 
  a pattern-variable in an action trigger, it tries to match any 
  pattern against it, and saves any match it finds so you can 
  use it in the response.  See below for examples.

  Note: As a note, actions are matched via regular expressions.
  %1 gets translated to (.+?) and %_1 gets translated to (\S+?).
  The special variable "%a" means "the whole matched line".

  We handle regular expressions with a special r[ ... ] syntax.  If
  you put an "i" or "I" after the ], then we'll ignorecase as well.

  The onetime argument can be set to true to have the action remove
  itself automatically after it is triggered.

  examples:
    #action {^You are hungry} {get bread bag;eat bread}
    #action {%0 gives you %5} {say thanks for the %5, %0!}
    #action {r[^%_1 tells\\s+you %2$]} {say %1 just told me %2}
    #action {r[sven dealt .+? to %1$]i} {say i just killed %1!}

  see also: unaction, enable, disable, atags
  
  category: commands
  """
  trigger = args["trigger"]
  action = args["action"]
  color = args["color"]
  priority = args["priority"]
  onetime = args["onetime"]
  quiet = args["quiet"]
  tag = args["tag"]

  am = exported.get_manager("action")
  ad = am.getActionData(ses)

  # they typed '#action'--print out all the current actions
  if not action:
    data = ad.getInfo(trigger, tag)
    if not data:
      data = ["action: no actions defined."]

    message = "actions"
    if tag:
      message += " with tag={%s}" % tag
      data += ad.getDisabledInfo(tag)
    exported.write_message(message + "\n" + "\n".join(data), ses)
    return

  try:
    ad.addAction(trigger, action, color, priority, onetime, tag)
    if not quiet:
      exported.write_message("action: {%s} {%s} color={%d} priority={%d} tag={%s} added." % (trigger, action, color, priority, str(tag)), ses)
  except:
    exported.write_traceback("action: exception thrown.", ses)

commands_dict["action"] = (action_cmd, "trigger= action= tag= color:boolean=false priority:int=5 onetime:boolean=false quiet:boolean=false")

def unaction_cmd(ses, args, input):
  """
  Removes action(s) from the manager.

  examples:
    #unaction {missed you.}
    #unaction missed*
    #unaction tag={indoor}
    
  see also: action, enable, disable, atags

  category: commands
  """
  am = exported.get_manager("action")
  ad = am.getActionData(ses)
  func = lambda x: ad.removeActions(x, args["tag"])
  modutils.unsomething_helper(args, func, None, "action", "actions")

commands_dict["unaction"] = (unaction_cmd, "str= tag= quiet:boolean=false")


def action_enable_cmd(ses, args, input):
  """
  Enables actions with given tag.
  By default, all the tags are enabled.
  
  see also: action, unaction, disable, atags

  category: commands
  """
  tag = args["tag"]
  am = exported.get_manager("action")
  ad = am.getActionData(ses)
  ad.enable(tag)

  if not args["quiet"]:
    exported.write_message("Enabling actions tagged as {%s}" % tag)
    
commands_dict["enable"] = (action_enable_cmd, "tag= quiet:boolean=false")

def action_disable_cmd(ses, args, input):
  """
  Temporarily disables all the actions with given tag, so their triggers
  won't trigger any actions (well, this desciption is a bit obscure,
  but I've tried my best :)

  see also: action, unaction, enable, atags
  
  category: commands
  """
  tag = args["tag"]
  am = exported.get_manager("action")
  ad = am.getActionData(ses)

  ad.disable(tag)

  if not args["quiet"]:
    exported.write_message("Disabling actions tagged as {%s}" % tag)
  
commands_dict["disable"] = (action_disable_cmd, "tag= quiet:boolean=false")

def action_tags_cmd(ses, args, input):
  """
  Shows all the tags available

  see also: action, unaction, enable, disable
  
  category: commands
  """
  list = exported.get_manager("action").listTags(ses)
  if list:
    exported.write_message("\n".join(list))
  else:
    exported.write_message("No tags defined.")

commands_dict["atags"] = (action_tags_cmd, "")  


am = None

def load():
  """ Initializes the module by binding all the commands."""
  global am, var_module
  modutils.load_commands(commands_dict)
  am = ActionManager()
  exported.add_manager("action", am)

  exported.hook_register("mud_filter_hook", am.mudfilter, 75)
  exported.hook_register("write_hook", am.persist)
  exported.hook_register("variable_change_hook", am.variableChange)

  from lyntin import config
  for mem in exported.get_active_sessions():
    # we need a separate BoolConfig for each session
    tc = config.BoolConfig("ignoreactions", 0, 1,
         "Allows you to turn off action handling.")
    exported.add_config("ignoreactions", tc, mem)

def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global am, var_module
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("alias")

  exported.hook_unregister("mud_filter_hook", am.mudfilter)
  exported.hook_unregister("write_hook", am.persist)
  exported.hook_unregister("variable_change_hook", am.variableChange)

  # remove configuration items for every session involved
  for mem in exported.get_active_sessions():
    exported.remove_config("ignoreactions", mem)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: logger.py,v 1.9 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines the LoggerManager which handles logging.

Logging can be turned on and shut off on a session by session basis.
"""
import string, os, threading
from lyntin import ansi, manager, config, utils, exported, constants
from lyntin.modules import modutils


class LoggerData:
  def __init__(self, session):
    self._logfile = None
    self._session = session
    # whether or not to strip ansi--0 is off, 1 is on
    self._strip_ansi = 0
    # pending user inputs:
    self._user_input = []
    # pending mud prompt:
    self._prompt = None
    self._userprefix = ''

    self._lock = thread.allocate_lock()

  def log(self, input):
    """
    Logs text to a file instance self._logfile and optionally
    filters ansi according to self._strip_ansi.

    @param input: the string to log to the logfile for this session
    @type  input: string
    """
    if self._logfile == None:
      return

    try:
      if self._strip_ansi == 1:
        input = ansi.filter_ansi(input)

      text = utils.filter_cm(input)
      text = text.replace("\n", os.linesep)
      self._logfile.write(text)
      self._logfile.flush()
    except:
      self._logfile = None
      exported.write_traceback("Logfile cannot be written to.", self._session)

  def log_mud(self, input):
    """
    Logs mud output, synchronizing it with user inputs.

    @param input: the string from the mud for this session
    @type  input: string
    """
    try:
      self._lock.acquire()
      
      if not input.endswith("\n"):
        # this is a prompt
        if self._prompt:
          # we already have one pending
          # (I'm not sure if it is possible, but who knows all the muds :)
          self.log(self._prompt[0]+"\n")
          self._prompt = (input, )
        elif self._user_input:
          # we have pending user input:
          self._log_user_input(input)
        else:
          # don't log prompts immediately, 
          # wait for a user input or a mud output
          self._prompt = (input, )
      else:
        # it is ordinal output from mud
        if self._prompt:
          # we have a prompt pending, let's log it first:
          self.log(self._prompt[0]+"\n")
          self._prompt = None
        self.log(input)
    finally:
      self._lock.release()

  def _log_user_input(self, prompt):
    self.log(''.join((prompt, "; ".join(self._user_input), "\n")))
    self._user_input = []

  def log_user(self, input):
    """
    Logs user input, synchronizing it with mud output.

    @param input: the string from user
    @type  input: string
    """
    try:
      self._lock.acquire()
      self._user_input.append(input)
      if self._prompt:
        # There is a prompt pending
        self._log_user_input(self._prompt[0])
        self._prompt = None
      elif self._userprefix: 
        # We have a default prefix, so we won't wait for next prompt;
        # instead we'll log all the user input we have right away
        # prepending it with the default prefix
        self._log_user_input(self._userprefix)
      elif len(self._user_input) > 100:
        # XXX something wrong, the server doesn't send prompts.
        # XXX Let's dump all the input we have into log now.
        # XXX For such a server one should use
        # XXX   #log <filename> userprefix="something"
        self._log_user_input(self._userprefix)
    finally:
      self._lock.release()

  def isLogging(self):
    """
    Tells whether we're logging or not.

    @returns: 1 if we're logging, 0 if not
    @rtype: boolean
    """
    if self._logfile == None:
      return 0
    return 1

  def closeLogFile(self):
    """
    Closes the logfile if it's currently open.
    """
    
    try:
      self._lock.acquire()
      if self._prompt:
        self.log(self._prompt[0]+"\n")
        self._prompt = None
      elif self._user_input:
        self._log_user_input(self._userprefix)
    finally:
      self._lock.release()

    if self._logfile:
      self._logfile.close()
      self._logfile = None

  def openLogFile(self, filename, stripansi=1, userprefix=''):
    """
    Opens a new logfile.

    @param filename: the name of the new file to open in append mode.
    @type  filename: string

    @param stripansi: whether (1) or not (0) to strip ansi from the
        logs
    @type  stripansi: boolean
    """
      
    # FIXME - what happens if we already have a logfile open?
    self.setLogFile(open(filename, "a"), stripansi, userprefix)

  def setLogFile(self, fileob, stripansi=1, userprefix=""):
    """
    Sets the logfile.

    @param fileob: the new File instance
    @type  fileob: File
    """
    self._logfile = fileob
    self._strip_ansi = stripansi
    self._userprefix = userprefix

  def clear(self):
    """
    Stops the logger.
    """
    self.closeLogFile()

  def getStatus(self):
    """
    Returns a one-liner describing this data object

    @return: one liner describing this object
    @rtype: string
    """
    if self._logfile:
      if self._strip_ansi == 1:
        return "logging to '" + self._logfile.name + "' (noansi)"
      else:
        return "logging to '%s'" % self._logfile.name
    else:
      return "logging not enabled"


class LoggerManager(manager.Manager):
  def __init__(self):
    self._loggers = {}

  def clear(self, ses):
    if ses in self._loggers:
      self._loggers[ses].clear()

  def getStatus(self, ses):
    if ses in self._loggers:
      return self._loggers[ses].getStatus()
    return "logging not enabled"

  def removeSession(self, ses):
    if ses in self._loggers:
      self._loggers[ses].closeLogFile()
      del self._loggers[ses]

  def getLogData(self, ses):
    if ses in self._loggers:
      return self._loggers[ses]

    logger = LoggerData(ses)
    self._loggers[ses] = logger
    return logger

  def mudfilter(self, args):
    """
    mud_filter_hook function for filtering incoming data from the mud.
    """
    ses = args["session"]
    text = args["dataadj"]

    logger = self._loggers.get(ses)
    if logger:
      logger.log_mud(text)

    return text

  def promptfilter(self, args):
    """
    prompt_filter_hook function for filtering incoming prompt from the mud.
    """
    logger = self._loggers.get(args["session"])
    text = args["prompt"]
    if logger:
      logger.log_mud(text)

    return text

  def tomudfilter(self, args):
    """
    to_mud_hook function for logging user input.
    """
    session = args["session"]
    text = args["data"]

    logger = self._loggers.get(session)
    if logger:
      logger.log_user(text)

    return text  


commands_dict = {}

def log_cmd(ses, args, input):
  """
  Will start or stop logging to a given filename for that session.
  Each session can have its own logfile.

  If USERPREFIX is set, then every line from the user will be 
  prepended with this prefix and immediately written into log file. 
  If USERPREFIX is omitted, then the user input will be attached to 
  mud prompts before logging.

  category: commands
  """
  logfile = args["logfile"]
  databuffer = args["databuffer"]
  stripansi = args["stripansi"]
  userprefix = args["userprefix"]

  if not ses.isConnected():
    exported.write_error("log: You must have a session to log.", ses)
    return

  lm = exported.get_manager("logger")
  loggerdata = lm.getLogData(ses)

  if not logfile:
    exported.write_message(loggerdata.getStatus(), ses)
    return

  # handle stopping logging
  if loggerdata.isLogging() == 1:
    try:
      logname = loggerdata._logfile.name
      loggerdata.closeLogFile()
      exported.write_message("log: stopped logging to '%s'." % logname, ses)
    except Exception as e:
      exported.write_error("log: logfile cannot be closed (%s)." % (e), ses)
    return

  # handle starting logging
  try:
    if os.sep not in logfile:
      logfile = config.options["datadir"] + logfile

    if databuffer:
      f = open(logfile, "w")
      buffer = "".join(ses.getDataBuffer())
      f.write(buffer)
      exported.write_message("log: dumped %d lines of databuffer to logfile" % buffer.count("\n"), ses)
      loggerdata.setLogFile(f, stripansi, userprefix)

    else:
      loggerdata.openLogFile(logfile, stripansi, userprefix)
    if stripansi:
      stripansimessage = " stripping ansi"
    else:
      stripansimessage = ""

    exported.write_message("log: starting logging to '%s'%s." % (logfile, stripansimessage), ses)
  except Exception as e:
    exported.write_error("log: logfile cannot be opened for appending. %s" % (e), ses)

commands_dict["log"] = (log_cmd, 'logfile= databuffer:boolean=false stripansi:boolean=true userprefix=')



lm = None

def load():
  """ Initializes the module by binding all the commands."""
  global lm
  modutils.load_commands(commands_dict)
  lm = LoggerManager()
  exported.add_manager("logger", lm)

  exported.hook_register("to_mud_hook", lm.tomudfilter, constants.LAST+1)
  exported.hook_register("mud_filter_hook", lm.mudfilter, 30)
  exported.hook_register("prompt_hook", lm.promptfilter, 30)

def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global lm
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("logger")

  exported.hook_unregister("to_mud_hook", lm.tomudfilter)
  exported.hook_unregister("mud_filter_hook", lm.mudfilter)
  exported.hook_unregister("prompt_hook", lm.promptfilter)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: testmodule.py,v 1.2 2007/07/24 00:39:03 willhelm Exp $
#########################################################################

import os, glob
import exported, modules.modutils, utils

index = __file__.rfind(os.sep)
if index == -1:
  path = "." + os.sep
else:
  path = __file__[:index]

if not path.endswith(os.sep):
  path += os.sep

commands_dict = {}

buffer = []

def capture_mud_filter(args):
  """
  Captures mud output.
  """
  buffer.append(str(args[1]) + "\n")

def capture_user_filter(args):
  """
  Captures ui output.
  """
  buffer.append(str(args[0]))

  
def runtest_cmd(ses, args, input):
  """
  Runs a test.

  Tests are essentially a commands file with some additional testing
  commands for the tester to handle.  The tester sits on the ui
  hook and copies all the data from when it begins to when it
  ends.  Then it takes this data and compares it against the benchmark
  (or if there is not benchmark, it creates one).  A test is successful
  if the benchmark equals the results of the current test run.

  It's pretty primitive--but it will work for most things.
  """
  global buffer

  test = args["test"]

  if not test:
    exported.write_message("available tests in %s:" % path)
    listing = glob.glob(os.path.join(path, "*.test"))
    listing = [mem.replace(path, "").replace(".test", "") for mem in listing]
    listing.sort()
    exported.write_message(utils.columnize(listing, indent=3))
    return

  exported.write_message("test started.")
  try:
    f = open (path + test + ".test", "r")
    lines = f.readlines()
    f.close()
  except:
    exported.write_traceback("runtest: cannot open test file %s." % test)
    return

  # bind ourselves to the hook so we capture all the data
  buffer = []
  exported.hook_register("to_user_hook", capture_user_filter)
  exported.hook_register("to_mud_hook", capture_mud_filter)

  for mem in lines:
    exported.write_message(">>> " + mem)
    exported.lyntin_command(mem, session=ses)

  exported.hook_unregister("to_user_hook", capture_user_filter)
  exported.hook_unregister("to_mud_hook", capture_mud_filter)

  results = "".join(buffer)

  try:
    f = open(path + test + ".benchmark", "r")
    benchmark = "".join(f.readlines())
    f.close()
  except:
    exported.write_traceback("runtest: cannot open benchmark--trying to save a new one.")
    try:
      f = open(path + test + ".benchmark", "w")
      f.write(results)
      f.close()
      exported.write_message("runtest: benchmark saved.")
    except:
      exported.write_traceback("runtest: could not save benchmark either.")
    return

  import difflib
  d = difflib.Differ()
  exported.write_message("----\n")
  exported.write_message("----\n")
  exported.write_message("diffing....")

  cmpresult = list(d.compare(benchmark.splitlines(1), results.splitlines(1)))
  cmpresult2 = [ mem[2:] for mem in cmpresult ]

  if cmpresult2 != benchmark.splitlines(1):
    exported.write_message("".join(cmpresult))
    exported.write_message("test failed.")
  else:
    exported.write_message("test passed.")

commands_dict["runtest"] = (runtest_cmd, "test=")

def load():
  modules.modutils.load_commands(commands_dict)


def unload():
  modules.modutils.unload_commands(commands_dict)



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: config.py,v 1.8 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module holds the configuration manager as well as a series of
configuration type classes.  It also holds some global variables
that get computed at boot.

X{config_change_hook}::

   This hook notifies registered functions that the value of a 
   config item has just been changed.

   Arg mapping: { "session": Session, "name": string, "oldvalue": string, "newvalue": string }

   session - The session of the config item that got changed.  If the
             session is None, then this is a global config item.

   name - the name of the config item

   oldvalue - the old value of the config item or None if there was
             no previous value.

   newvalue - the new value of the config item
"""
import types, copy
from lyntin import exported, utils, manager, constants, config

# this holds a list of all the modules Lyntin has dynamically imported
# or have been imported via the #import command.
lyntinmodules = []

# holds the application options--these are adjusted by command-line 
# arguments only
options = {'datadir': '',
           'moduledir': [],
           'readfile': [],
           'snoopdefault': 1,
           'ui': 'text'}

class ConfigBase:
  """
  A ConfigDataBase subclass encapsulates a piece of configuration.  It
  encapsulates its type functionality (data checking, et al), its name
  and its value.
  """
  def __init__(self, name, originalvalue, persist, description):
    """
    Sets the name and original value.  Override this if you need
    additional pieces of information in your ConfigDataBase
    subclass.

    @param name: the name of the config item
    @type  name: string

    @param originalvalue: the original value of the config item
    @type  originalvalue: varies

    @param persist: this config item should (1) or should not (0) persist
        via the write command
    @type  persist: boolean

    @param description: a short explanation of what this config item
        is.
    @type  description: string
    """
    self._name = name
    self._value = originalvalue
    self._persist = persist
    self._description = description

  def set(self, newvalue):
    """
    Sets the value.  This should check the value first, and then
    set it if it's ok.

    Override this if you need to provide additional functionality.

    @param newvalue: the new value to set
    @type  newvalue: varies

    @returns: the old value (actual value)
    @rtype: varies
    """
    oldvalue = self._value

    # check should raise an Exception if the value is not appropriate.
    # it should also return the adjusted value (string -> whatever we're
    # supposed to store)
    self._value = self.check(newvalue)

    return oldvalue

  def check(self, value):
    """
    Checks the value to verify that it's a valid value for this
    config data item.  Override this so that it checks the incoming
    value and if it's not a valid type, it kicks up a TypeError.
    If it's not a valid value (different than not being a valid 
    type) it kicks up a ValueError.  If everything is ok, it should
    return the converted argument.

    @param value: the value to check (and convert)
    @type  value: string

    @returns: the converted value if everything is kosher
    @rtype: varies

    @raises TypeError: if it's not a valid type
    @raises ValueError: if it's not a valid value
    """
    return value

  def get(self):
    """
    Retrieves the value in question.

    @returns: the value we're holding
    @rtype: varies
    """
    return self._value

  def toString(self):
    """
    Retrieves a textual representation of the value and the type
    of the config item.

    @returns: string
    @rtype: string
    """
    return repr(self._value) + " (varies)"

  def getDescription(self):
    """
    Gets the description of the config item.

    @returns: the description of the config item
    @rtype: string
    """
    return self._description


class StringConfig(ConfigBase):
  """
  Holds a string.
  """
  def check(self, value):
    if not isinstance(value, str):
      raise TypeError("Value must be of type str.")
    return value

  def toString(self):
    return repr(self._value) + " (string)"

class CharConfig(ConfigBase):
  """
  Holds a single character.
  """
  def check(self, value):
    """
    Ensures the value is a single character string.

    :param value: The value to check.
    :return: The validated value if it's a single character string.
    :raises TypeError: If the value is not a string.
    :raises ValueError: If the string is not exactly one character long.
    """
    if not isinstance(value, str):
      raise TypeError("Value is not of type string.")

    if len(value) != 1:
      raise ValueError("Value must be only one character.")

    return value

  def toString(self):
    return repr(self._value) + " (char)"


class IntConfig(ConfigBase):
  """
  Holds an int.
  """
  def check(self, value):
    return int(value)
  
  def toString(self):
    return repr(self._value) + " (int)"

def bv(bool):
  if bool:
    return "on"
  return "off"

class BoolConfig(ConfigBase):
  """
  Holds a boolean value.
  """
  def check(self, value):
    ret = utils.convert_boolean(value)
    if ret == 1 or ret == 0:
      return ret

    raise ValueError("Invalid boolean value specified: %s" % (value))

  def toString(self):
    return bv(self._value) + " (bool)"


class ConfigManager(manager.Manager):
  """
  Holds all the configuration pieces for Lyntin.
  """
  def __init__(self, e):
    # this is a map of session -> (map of config names -> items)
    self._config = {}
    self._engine = e

  def add(self, name, configitem, ses=None):
    """
    Adds a new configuration item.

    @param name: the name of the item
    @type  name: string

    @param configitem: the configuration item to add
    @type  configitem: ConfigBase

    @param ses: if this item is session based, then this is the session
        to associate the item with
    @type  ses: Session

    @raises ValueError: if we already have an item in that session with
        that name.
    """
    if ses not in self._config:
      self._config[ses] = {}

    if name in self._config[ses]:
      raise ValueError("Already have a config item of that name.")

    self._config[ses][name] = configitem
    self._configChangeHook(ses, name, None, configitem.get())

  def remove(self, name, ses=None):
    """
    Allows you to remove a configuration item from the system.

    @param name: the name of the item to remove
    @type  name: string

    @param ses: the session from which to remove the item (None if
        it's a general Lyntin item)
    @type  ses: Session

    @raises ValueError: if the item does not exist
    """
    if ses not in self._config:
      raise ValueError("That session does not exist.")

    if name not in self._config[ses]:
      raise ValueError("That item does not exist.")

    del self._config[ses][name]
    
  def change(self, name, newvalue, ses=None):
    """
    Changes the value of a configuration item and then (if 
    successful), it spams the config_change_hook with the
    session (or None), name of the config item, the old value 
    and the new value.

    @param name: the name of the item to change
    @type  name: string

    @param newvalue: the new value to change it to
    @type  newvalue: varies (probably a string)

    @param ses: the session (or None if this is not session-scoped)
    @type  ses: Session
    """
    if ses not in self._config:
      raise ValueError("Session '%s' does not exist." % repr(ses))

    if name not in self._config[ses]:
      raise ValueError("No config item of that name.")

    oldvalue = self._config[ses][name].set(newvalue)
    self._configChangeHook(ses, name, oldvalue, self._config[ses][name].get())

  def get(self, name, ses=None, defaultvalue=constants.NODEFAULTVALUE):
    """
    Gets a value for a config item.  If the default value is
    not specified, then it will raise a ValueError.

    @param name: the name of the item to retrieve the value of
    @type  name: string

    @param ses: the session (or None if this is not session-scoped)
    @type  ses: Session
     
    @param defaultvalue: the value to return if there is no config
        item of that name.  if you don't specify this, then we'll
        raise a ValueError.
    @type  defaultvalue: varies
    """
    if ses not in self._config:
      if ses is None:
        self._config[None] = {}
      else:
        raise ValueError("Session '%s' does not exist." % repr(ses))

    if name not in self._config[ses]:
      if defaultvalue == constants.NODEFAULTVALUE:
        raise ValueError("No config item of that name")
      else:
        return defaultvalue

    return self._config[ses][name].get()

  def _configChangeHook(self, ses, name, value, newvalue):
    exported.hook_spam("config_change_hook", 
        {"session": ses, "name": name, "oldvalue": value, "newvalue": newvalue })

  def getConfigItems(self, ses=None):
    """
    Returns all the configuration items for a given session.

    @param ses: the session to pull items for, or None if it's general
        Lyntin configuration stuff
    @type  ses: Session

    @returns: list of ConfigBase subclass objects
    @rtype: list of ConfigBase subclass objects
    """
    if ses in self._config:
      return self._config[ses].values()

    return []

  def getConfigItem(self, name, ses=None):
    """
    Retrieves a specific configuration item by name and session.  If
    the ses passed in is None, then we look for lyntin general config
    items.

    @param name: the name of the config item to retrieve
    @type  name: string

    @param ses: the session to retrieve the item from (or None if it's
        a global item)
    @type  ses: Session

    @returns: a ConfigBase item or None if the item doesn't exist
    @rtype: ConfigBase
    """
    if ses not in self._config:
      return None

    if name not in self._config[ses]:
      return None

    return self._config[ses][name]


  def addSession(self, newsession, basesession):
    # if we have nothing to clone from, then we don't want to
    # worry about this
    if not basesession or basesession not in self._config:
      return

    x = {}

    for mem in list(self._config[basesession].values()):
      x[mem._name] = copy.deepcopy(mem)

    self._config[newsession] = x

  def removeSession(self, ses):
    if ses in self._config:
      del self._config[ses]




#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: __init__.py,v 1.3 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
The modules package holds all of the dynamically loaded Lyntin modules.
Modules get loaded when Lyntin starts up unless:

  1. the module throws an exception when getting imported
  2. the module's name starts with an _

On multi-user systems, you'll want to dump modules that everyone will want
to use here.  Otherwise, users can put modules that they want to use in
their moduledir and specify the moduledir at the command line using the
-m flag.
"""

import glob, os, sys
from lyntin import exported, config


def test_for_conflicts(name, module):
  """
  Tests a module we just imported with the name the path the module
  should have.  This allows us to test Lyntin modules we just dynamically
  loaded to verify it's the one we intended to load.

  Right now we don't really do anything except kick up an error to the
  user.  Let them deal with the issue.

  @param name: the full name of the module we wanted to load
  @type  name: string

  @param module: the actual module we loaded
  @type  module: module instance
  """
  if module.__file__ != name + "c" and module.__file__ != name:
    exported.write_error("possible name conflict: '%s' and '%s'" % 
                         (name, module.__file__))


def get_module_name(filename):
  """
  Takes in a fully qualified filename and returns the module name
  portion.

  example::

    /home/willg/lyntinng/modules/alias.py -> alias

  @param filename:  the fully qualified filename
  @type filename: string

  @returns: the module name
  @rtype: string
  """
  path, filename = os.path.split(filename)
  return os.path.splitext(filename)[0]
  
def load_modules():
  """
  Magically dynamically loads all the modules in the modules
  package.  This is truly a semi-magic function.
  """
  # handle modules.*
  index = __file__.rfind(os.sep)
  if index == -1:
    path = "." + os.sep
  else:
    path = __file__[:index]

  _module_list = glob.glob( os.path.join(path, "*.py"))
  _module_list.sort()

  for mem in _module_list:
    # we skip over all files that start with a _
    # this allows hackers to be working on a module and not have
    # it die every time.
    mem2 = get_module_name(mem)
    if mem2.startswith("_"):
      continue

    try:
      name = "lyntin.modules." + mem2
      _module = __import__(name)
      _module = sys.modules[name]

      if "load" in _module.__dict__:
        _module.load()

      _module.__dict__["lyntin_import"] = 1
      config.lyntinmodules.append(name)
    except:
      exported.write_traceback("Module '%s' refuses to load." % name)

  # handle modules found in the moduledir
  moduledirlist = config.options["moduledir"]
  if moduledirlist:
    for moduledir in moduledirlist:
      # grab the contents of the moduledir directory
      _module_list = glob.glob( os.path.join( moduledir, "*.py"))

      # toss the moduledir in the sys.path
      sys.path.append(moduledir)

      # and toss all the contents of the directory in our _module_list
      for mem in _module_list:
        mem2 = get_module_name(mem)
        if mem2.startswith("_"):
          continue

        try:
          _module = __import__(mem2)
          if "load" in _module.__dict__:
            _module.load()

          test_for_conflicts(mem, _module)

          _module.__dict__["lyntin_import"] = 1
          config.lyntinmodules.append(mem2)
        except:
          exported.write_traceback("Module '%s' refuses to load." % mem)


# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: modutils.py,v 1.4 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module holds helper functions for building other Lyntin modules.
This module will likely make things easier for you, however, it is not
an API module and the contents herein are subject to change if we
need to change them.  Having said that, I will note it doesn't change
much.
"""
import types
from lyntin import exported


def load_commands(commands_dict):
  """
  Takes in a dict and loads all the commands in that dict.

  The dict is a mapping of command name to arguments to be passed
  to exported.add_command.  Pretty much we just turn around and
  call exported.add_command with the arguments in the dict without
  any transformation at all.

  @param commands_dict: the map holding the command names and the
      arguments we need to call exported.add_command repeatedly
  @type  commands_dict: dict
  """
  for mem in list(commands_dict.keys()):
    args = commands_dict[mem]
    if isinstance(args, tuple):
      exported.add_command(*((mem,)+args))
    else:
      exported.add_command(mem, args)

def unload_commands(commands_list):
  """
  Takes in a list of command names and removes the commands from
  Lyntin by calling exported.remove_command.

  @param commands_list: the list of command names to remove
  @type  commands_list: list of strings
  """
  for mem in commands_list:
    exported.remove_command(mem)


def unsomething_helper(args, func, ses=None, sing="", plur=""):
  """
  Helps automate some of the un(something) commands.  These are
  commands that remove data from a given manager.  For example,
  unalias, unaction....

  This method also handles printing out status information to
  the user as to how many somethings were removed.

  @param args: the map with the str and quiet arguments in it
  @type  args: dict

  @param func: the function to call to actually do the work.
      it should take in a single string argument.
  @type  func: function

  @param ses: the session to apply this to
  @type  ses: Session

  @param sing: the singular form of the unsomething--for output
      (ex: "alias")
  @type  sing: string

  @param plur: the plural form of the unsomething--for output
      (ex: "aliases")
  @type  plur: string
  """
  text = args["str"]
  quiet = args["quiet"]

  if ses:
    removedthings = func(ses, text)
  else:
    removedthings = func(text)

  if not quiet:
    if len(removedthings) == 0:
      data = "un%s: No %s removed." % (sing, plur)
    else:
      data = []
      for mem in removedthings:
        if isinstance(mem, (tuple, list)):
          mem = "{%s}" % ("} {".join(map(str, mem)))
          data.append("un%s: %s removed." % (sing, mem))
        else:
          data.append("un%s: {%s} removed." % (sing, mem))
      data = "\n".join(data)
    exported.write_message(data, ses)


# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#!/usr/bin/env python
#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: testserver.py,v 1.2 2005/01/07 14:50:40 glasssnake Exp $
#######################################################################
"""
This new test-server is a patchwork of stuff from the existing test server
and code I wrote for the Varium mud server way back when.  It is actually
a functional mini-mud now.
"""
import socket, sys, queue, select, textwrap

import connection

my_world = None

class Event:
  def __init__(self):
    pass

  def execute(self, world):
    pass

  def __str__(self):
    return ""

class InputEvent(Event):
  def __init__(self, conn, input):
    super().__init__()
    self._conn = conn
    self._input = input

  def __str__(self):
    return f'"{self._input}" for "{self._conn}"'

  def execute(self, world):
    self._conn.handleInput(world, self._input)

class HeartbeatEvent(Event):
  def __init__(self, source):
    super().__init__()
    self._source = source

  def execute(self, world):
    self._source.heartbeat(world)

class NPC:
  def __init__(self, world):
    self._world = world
    self._name = "Joe"
    self._desc = "A regular looking NPC."

    from random import Random
    self._random = Random()

  def blab(self):
    self._world.spamroom(self._name + " looks fidgety.\n")

class Neil(NPC):
  def __init__(self, world):
    NPC.__init__(self, world)
    self._name = "Neil"
    self._desc = "Neil is the bartender at this little mini-tavern."

  def heartbeat(self, world=None):
    g = (self._random.random() * 10)
    if (g < 2):
      self._world.spamroom(self._name + " flicks a bug off his bar.\n")
    elif (g < 4):
      self._world.spamroom(self._name + " scrubs some glasses with his dishrag.\n")
    elif (g < 5):
      self._world.spamroom(self._name + " says, \"Mighty fine morning, isn't it?\"\n")

class World:
  def __init__(self, options):
    self._event_queue = queue.Queue(0)
    self._worker = None
    self._options = options
    self._ms = None

    temp = ("Welcome to Neil's Pub--a small tavern of unusual candor.  " +
            "In many ways, this is a dream come true for Neil and it shows " +
            "in the care he gives to the place.  The tavern is both " +
            "infinitely large and terribly small.  There are no exits.  " +
            "Only a long bar and a series of barstools for folks to show " +
            "up, take a load off, and socialize.")

    self._desc = textwrap.fill(temp, width=70)

    self._npcs = []

    self._npcs.append(Neil(self))

  def enqueue(self, ev):
    self._event_queue.put(ev)

  def startup(self):
    from threading import current_thread
    self._worker = current_thread()

    # launch the server
    self._ms = MasterServer(self, self._options)
    self._ms.startup()

    do_heartbeat = self._options.get("heartbeat")
    beat = 0

    # this is our main loop thingy!
    while True:
      if do_heartbeat == "yes":
        beat += 1
        if beat % 30 == 0:
          beat = 0
          self.heartbeat()

      self._ms.networkLoop()

      # Handle events
      try:
        event = self._event_queue.get(block=False)
        es = str(event)
        event.execute(self)
      except queue.Empty:
        pass  # No events to process
      except Exception as e:
        print(f"Exception while executing event: {e}")

  def heartbeat(self):
    """Triggers NPC heartbeats when players are connected."""
    if self._ms and hasattr(self._ms, "_conns") and len(self._ms._conns) > 1:
      for npc in self._npcs:
        self.enqueue(HeartbeatEvent(npc))

  def disconnect(self, conn):
    """Handles player disconnection."""
    print("DEBUG: 'disconnect' function has been called.")
    if self._ms and hasattr(self._ms, "_conns") and conn in self._ms._conns:
      print(f"About to close and remove: {self._ms._conns(conn)}")
      self._ms._conns.remove(conn)
      self._ms._conns.close()

    if hasattr(conn, '_name'):
      self.spamroom(f"{conn._name} has left the game.\n")

    print(f"Goodbye {getattr(conn, '_name', 'Unknown Player')}")

  def look(self, conn, item):
    if item:
      for mem in self._npcs:
        if item == mem._name.lower():
          return mem._desc + "\n"

      for mem in self._ms._conns:
        if item == mem._name.lower():
          return mem._desc + "\n"

    else:
      out = self._desc + "\n\n"

      # Convert the map object to a list
      names = [x._name for x in self._ms._conns]

      # Append additional names from self._npcs
      names.extend(mem._name for mem in self._npcs)

      # Join the names with ', ' as a separator
      out += textwrap.fill(', '.join(names), 70)

      out += "\n"

      return out

    return "That does not exist.\n"

  def spamroom(self, data):
    for mem in self._ms._conns:
      try:
        if hasattr(mem, 'write'):
          mem.write(data) #connection still alive here.
        else:
          print(f"Skipping invalid connection: {mem}")
      except Exception as e:
        print(f"Error writing to {mem}: {e}")

  def shutdown(self):
    if self._ms:
      self._ms.closedown()
    print("Shutting down.")

class MasterServer:
  def __init__(self, world, options):
    super().__init__()
    self._args = args
    self._master = None
    self._conns = []
    self._shutdown = 0
    self._options = options
    self._world = world

  def startup(self):
    host = self._options["host"]
    port = int(self._options["port"])
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind((host, port))
    print(f"Socket bound to {host}:{port}")
    s.listen(5)
    print("Socket is listening for incoming connections...")

    print(f"Test server starting up: {host}:{port}")
    self._master = connection.Connection(self._world, s, "MASTER")
    self._master._name = "Igor"
    self._master._desc = "A very busy old man."

    self._conns = []
    self._conns.append(self._master)

  def networkLoop(self):
    # Prepare the list of socket objects from the connections
    #fns = [conn.sockid() for conn in self._conns]  # Use list comprehension instead of map
    fns = [conn.sockid() for conn in self._conns if conn.sockid() and hasattr(conn.sockid(), "fileno")]
    #print("DEBUG: fns =", fns)  # Should only contain socket objects!

    fi = select.select(fns, [], [], 0.1)[0]  # Get the readable sockets
    allconns = [conn for conn in self._conns if conn.sockid() in fi]  # Use list comprehension instead of filter

    for conn in allconns:
      if conn._addr == "MASTER":
        # Accept a new connection
        try:
          newsock, newaddr = conn._sock.accept()
          newconn = connection.Connection(self._world, newsock, newaddr)
          newconn.write("Welcome to Neil's Pub! Type \"help\" if you're lost.\n")
          self._conns.append(newconn)
        except Exception as e:
          print(f"Error accepting new connection: {e}")

      else:
        # Handle existing connections
        try:
          new_data = conn._sock.recv(1024)
          if new_data:
            new_data = new_data.decode('utf-8')
        except Exception as e:
          print(f"Exception: {e}")
          if conn in self._conns:
            self._conns.remove(conn)
          if conn._sock:  # Only call killConn() if the socket still exists
            conn.killConn()
          continue

        if new_data:
          conn.handleNetworkData(new_data)
        else:
          if conn in self._conns:
            #self._conns.remove(conn)
            conn.killConn()

  def closedown(self):
    print("DEBUG: Closing the socket.")
    #try:
      #if self._master.sockid() != -1:
          #self._master._sock.close()
    #except Exception as e:
      #print(f"Closing down master socket failed: '{e}'")

    for mem in self._conns[:]:  # Iterate over a copy of the list
      try:
        if mem._addr != "MASTER":
          mem.write("Server shutting down.\n")
        mem.killConn()
      except Exception as e:
        print(f"Error closing connection for {mem._addr}: {e}")

def print_syntax(message=""):
  print("testserver.py [--help] [options]")
  print("    -h|--host <hostname> - sets the hostname to bind to")
  print("    -p|--port <port>     - sets the port to bind to")
  print("    --heartbeat <yes|no> - sets whether or not to execute heartbeats")
  print()
  if message:
    print(message)

if __name__ == '__main__':
  import testserver, signal, traceback, logging

  # parse out arguments
  args = sys.argv[1:]
  i = 0
  optlist = []
  while (i < len(args)):

    if args[i][0] == "-":
      if (i+1 < len(args)):
        if args[i+1][0] != "-":
          optlist.append((args[i], args[i+1]))
          i = i + 1
        else:
          optlist.append((args[i], ""))
      else:
        optlist.append((args[i], ""))

    else:
      optlist.append(("", args[i]))

    i = i + 1

  options = {"host": "localhost", "port": "3000", "heartbeat":"yes"}
  print("Handling arguments.")
  for mem in optlist:
    if mem[0] == "--host" or mem[0] == "-h":
      if mem[1]:
        options["host"] = mem[1]
      else:
        print_syntax("error: Host was not specified.")
        sys.exit(1)

    elif mem[0] == "--help":
      print_syntax()
      sys.exit(1)

    elif mem[0] == "--port" or mem[0] == "-p":
      if mem[1] and mem[1].isdigit():
        options["port"] = mem[1]
      else:
        print_syntax("error: Port needs to be a number.")
        sys.exit(1)

    elif mem[0] == "--heartbeat":
      if mem[1].lower() == "yes" or mem[1].lower() == "no":
        options["heartbeat"] = mem[1].lower()
      else:
        print_syntax("error: Valid heartbeat settings are 'yes' or 'no'.")
        sys.exit(1)

  print(f"Host: {options['host']}")
  print(f"Port: { options['port']}")

  # Setup logging
  logging.basicConfig(level=logging.ERROR, filename="server_error.log", filemode="a",
                      format="%(asctime)s - %(levelname)s - %(message)s")

  # Create the world
  testserver.my_world = World(options)

  try:
    testserver.my_world.startup()
  except KeyboardInterrupt:
    print("Server interrupted by user. Shutting down...")
  except ValueError as ve:
    print(f"ValueError: {ve}")
    logging.error("ValueError occurred", exc_info=True)
  except OSError as oe:
    print(f"OS error: {oe}")
    logging.error("OS error occurred", exc_info=True)
  except Exception as e:
    print(f"Outer loop exception: {e}")
    logging.error("Unhandled exception", exc_info=True)
    traceback.print_exc()  # Print full stack trace to console
  finally:
    testserver.my_world.shutdown()
    print("Server shut down gracefully.")

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: alias.py,v 1.9 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines the AliasManager which manages aliases, creating new
aliases, removing aliases, checking user input for aliases, expanding
aliases, and other such things.

The AliasManager has an AliasData object for every session that has
aliases.

An alias consists of an "alias" and an "expansion".  So if there is
an alias:

   l3k -> #ses a localhost 3000

the "alias" is "l3k" and the "expansion" is "#ses a localhost 3000".
Whenever the user types "l3k" this module expands it to 
"#ses a localhost 3000" which then (after going through other user_filter
hook functions) gets executed.

Aliases are currently handled via string finding and not regular
expressions.  At some point in the future, this will be changed to
regular expressions to better handle a wider variety of aliases.
"""
from lyntin import manager, utils, exported
from lyntin.modules import modutils

class AliasData:
  """ Manages aliases."""
  def __init__(self):
    self._aliases = {}

  def addAlias(self, name, expansion):
    """
    Adds an alias to the dict.

    @param name: the alias name
    @type  name: string

    @param expansion: the string the alias expands to
    @type  expansion: string

    @raises ValueError: when the name is the same as the expansion
    """
    if name == expansion:
      raise ValueError("name cannot equal expansion.")
    self._aliases[name] = expansion

  def clear(self):
    """
    Removes all the aliases.
    """
    self._aliases.clear()

  def removeAliases(self, text):
    """
    Removes aliases from the list.

    Returns a list of tuples of alias name/expansion that
    were removed.

    @param text: the text which when run through util.expand
        gives us the aliases to remove
    @type  text: string

    @return: the list of alias/expansions that match the text
    @rtype: list of (string, string)
    """
    badaliases = utils.expand_text(text, list(self._aliases.keys()))

    ret = []
    for mem in badaliases:
      ret.append((mem, self._aliases[mem]))
      del self._aliases[mem]

    return ret

  def getAliases(self):
    """
    Returns the keys of the alias dict.

    @return: all the aliases we're managing (but not the expansions)
    @rtype: list of strings
    """
    listing = sorted(self._aliases.keys())
    return listing

  def expand(self, input):
    """
    Looks at user input and expands any aliases involved.

    It'll return the expansion if there is one.  Otherwise
    it returns None.

    @param input: the user input to expand
    @type  input: string

    @return: the alias expansion for the given input if it's an alias
        or None
    @rtype: string
    """
    if len(input) > 0:
      # pull out the first word of the input
      firstword = input.split(' ', 1)[0]

      # if we match an alias, we return the expansion
      if firstword in list(self._aliases.keys()):
        return self._aliases[firstword]            

    return None

  def getStatus(self):
    """
    Returns the one-line status of this manager.

    @return: the one line status
    @rtype: string
    """
    return "%d alias(es)." % len(self._aliases)
    
  def getInfo(self, text=""):
    """
    Returns information about the aliases in here.

    This is used by #alias to tell all the aliases involved
    as well as #write which takes this information and dumps
    it to the file.

    @param text: the text to expand to find aliases the user
        wants information about.
    @type  text: string

    @return: a list of strings where each string represents an alias
    @rtype: list of strings
    """
    if len(list(self._aliases)) == 0:
      return []

    listing = list(self._aliases.keys())
    if text:
      listing = utils.expand_text(text, listing)

    data = []
    for mem in listing:
      data.append("alias {%s} {%s}" % (mem, utils.escape(self._aliases[mem])))

    return data

  def getInfoMappings(self):
    l = []
    for m in list(self._aliases.keys()):
      l.append( { "alias": m, "expansion": self._aliases[m] } )

    return l


class AliasManager(manager.Manager):
  """
  Extends the manager.Manager base class to implement alias
  functionality.  The AliasManager holds session -> AliasData objects
  which handle the alias functionality.
  """
  def __init__(self):
    # session -> AliasData objects
    self._aliasdata = {}

  def getAliasData(self, ses):
    if ses not in self._aliasdata:
      self._aliasdata[ses] = AliasData()
    return self._aliasdata[ses]

  def clear(self, ses):
    if ses not in self._aliasdata:
      return
    self._aliasdata[ses].clear()

  def getAlias(self, ses, text):
    return self.getAliasData(ses).expand(text)

  def getStatus(self, ses):
    return self.getAliasData(ses).getStatus()

  def getInfo(self, ses, text=""):
    return self.getAliasData(ses).getInfo(text)

  def getInfoMappings(self, item, ses):
    if item != "alias":
      raise ValueError("%s is not a valid item for this manager." % item)

    if ses in self._aliasdata:
      return self._aliasdata[ses].getInfoMappings()
    return []
    
  def getItems(self):
    return [ "alias" ]

  def getParameters(self, item):
    if item != "alias":
      raise ValueError("%s is not a valid item for this manager." % item)

    return [ ("alias", "The alias name."),
             ("expansion", "The alias expansion.") ]

  def persist(self, args):
    """
    write_hook function for persisting the state of our session.
    """
    ses = args["session"]
    quiet = args["quiet"]

    data = self.getInfo(ses)
    if quiet == 1:
      data = [m + " quiet={true}" for m in data]
        
    return data

  def userfilter(self, args):
    """ 
    user_filter_hook.
    """
    # we check for aliases here--and if we find some, we
    # do the variable expansion and then recurse over the result
    ses = args["session"]
    internal = args["internal"]
    verbatim = args["verbatim"]
    text = args["dataadj"]
  
    if ses not in self._aliasdata or verbatim == 1:
      return text

    aliasexpansion = self._aliasdata[ses].expand(text)

    if not aliasexpansion:
      return text
    else:
      aliasexpansion = utils.expand_placement_vars(text, aliasexpansion)
      exported.lyntin_command(aliasexpansion, 1, ses)
      return None

  def addSession(self, newsession, basesession=None):
    if basesession:
      if basesession in list(self._aliasdata):
        bdata = self.getAliasData(basesession)
        ndata = self.getAliasData(newsession)

        for mem in list(bdata._aliases.keys()):
          ndata.addAlias(mem, bdata._aliases[mem])

  def removeSession(self, ses):
    if ses in self._aliasdata:
      del self._aliasdata[ses]


commands_dict = {}

def alias_cmd(ses, args, input):
  """
  With no arguments, prints all aliases.
  With one argument, prints all aliases which match the arg.
  With multiple arguments, creates an alias.

  You can use pattern variables which look like % and a number.  %0 
  will be all the arguments passed in.

  Ranges can be used by using python colon-syntax, specifying a
  half-open slice of the input items, so %0:3 is the alias name, first,
  and second arguments of the input.

  Negative numbers count back from the end of the list.  So %-1 is the
  last item in the list, %:-1 is everything but the last item in the
  list. 

  examples:
    #alias {k*}                    - prints out aliases that start with k
    #alias {k} {kill %1}           - builds a new alias
    #alias {gg} {put %1: in chest} - builds a new alias

  category: commands
  """
  name = args["alias"]
  command = args["expansion"]
  quiet = args["quiet"]

  am = exported.get_manager("alias")
  ad = am.getAliasData(ses)

  # they typed '#alias' or '#alias x' so we print the relevant aliases
  if not command:
    data = ad.getInfo(name)
    if not data:
      data = ["alias: no aliases defined."]

    exported.write_message("aliases:\n" + "\n".join(data), ses)
    return

  # they're creating an alias
  try:
    ad.addAlias(name, command)
  except ValueError as e:
    exported.write_error("alias: %s" % e, ses)

  if not quiet:
    exported.write_message("alias: {%s} {%s} added." % (name, command), ses)

commands_dict["alias"] = (alias_cmd, "alias= expansion= quiet:boolean=false")


def unalias_cmd(ses, args, input):
  """
  Allows you to remove aliases.

  category: commands
  """
  func = exported.get_manager("alias").getAliasData(ses).removeAliases
  modutils.unsomething_helper(args, func, None, "alias", "aliases")

commands_dict["unalias"] = (unalias_cmd, "str= quiet:boolean=false")

am = None

def load():
  """ Initializes the module by binding all the commands."""
  global am
  modutils.load_commands(commands_dict)
  am = AliasManager()
  exported.add_manager("alias", am)

  exported.hook_register("user_filter_hook", am.userfilter, 20)
  exported.hook_register("write_hook", am.persist)

def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global am
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("alias")

  exported.hook_register("user_filter_hook", am.userfilter)
  exported.hook_register("write_hook", am.persist)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: tintincmds.py,v 1.26 2007/07/24 00:39:03 willhelm Exp $
#########################################################################

import os, os.path
from lyntin import net, utils, engine, constants, config, exported, event
from lyntin.modules import modutils

"""
This module holds commands that are derived from Tintin, but don't involve
a manager.  Tintin commands that involve a manager (alias and unalias,
action and unaction, variable and unvariable...) are in their respective
modules along with their manager and any helper functions involved.
"""
commands_dict = {}


def bell_cmd(ses, words, input):
  """
  Kicks off the bell for a given session.  Anything registered
  with the bell_hook will get tickled.

  category: commands
  """
  exported.hook_spam("bell_hook", {"session": ses})

commands_dict["bell"] = (bell_cmd, "")


def clear_cmd(ses, words, input):
  """
  This command clears a session of all session data (except the actual 
  connection).  This covers gags, subs, actions, aliases...

  category: commands
  """
  try:
    ses.clear()
    exported.write_message("clear: session %s cleared." % ses.getName(), ses)
  except Exception as e:
    exported.write_traceback("clear: error in clearing session %s" % ses)

commands_dict["clear"] = (clear_cmd, "")
  

def cr_cmd(ses, args, input):
  """
  This sends a carriage return to the mud.  This is useful in aliases 
  and actions that require a carriage return.

  category: commands
  """
  ses.writeSocket("\n")

commands_dict["^cr"] = (cr_cmd, "")


def end_cmd(ses, args, input):
  """
  Closes all sessions and quits out of Lyntin.

  Note: on most muds this will leave your character in a state of 
  linkdeath--it does not sell all your stuff, return you to town, 
  save your character, tell your friends goodbye, or anything of 
  that nature.

  category: commands
  """
  exported.write_message("end: you'll be back...")
  event.ShutdownEvent().enqueue()

commands_dict["^end"] = (end_cmd, "")


def help_cmd(ses, args, input):
  """
  With no arguments, shows all the help files available.
  With an argument, shows that specific help file or lists the contents
  of that category of help files.

  examples:
    #help                      - lists all help files in the root
    #help help                 - shows help for the help command
    #help commands.substitute  - shows help for the substitute command
    #help commands             - shows help for the commands category

  Items that have a number in parentheses after them are a category.
  The number is how many help topics are below that category.

  example:
    > #help
    lyntin: ::Lyntin Help::
    lyntin:
    lyntin: category: root
    lyntin:
    lyntin:    commands(55)  readme(13)
    lyntin:    textui
    >

  category: commands
  """
  item = args["item"]

  keys = item.split(".")
  data = "::Lyntin Help::\n\n"

  error, breadcrumbs, text = exported.get_help(item)

  if error:
    data += "%s\n\n" % error
  if breadcrumbs:
    data += "category: %s\n\n" % breadcrumbs

  data += text
  exported.write_message(data)

commands_dict["help"] = (help_cmd, "item=")


def history_cmd(ses, args, input):
  """
  #history prints the current history buffer.

  ! will call an item in the history indexed by the number after
  the !.  You can also do replacements via the sub=repl syntax.

  examples:
    #history [count=30]
        prints the last count entries in the history buffer
    !
        executes the last thing you did
    !4
        executes the fourth to last thing you did
    !4 3k=gk
        executes the fourth to last thing you did after replacing
        3k with gk in it

  category: commands
  """
  count = args["count"]
  
  historylist = exported.get_history(count)
  for i in range(0, len(historylist)):
    historylist[i] = "%d %s" % ((i+1), historylist[i])
  historylist.reverse()
  exported.write_message("History:\n" + "\n".join(historylist))

commands_dict["history"] = (history_cmd, "count:int=30")


def if_cmd(ses, args, input):
  """
  Allows you to do some boolean logic based on Lyntin variables
  or any Python expression.  If this expression returns a non-false
  value, then the action will be performed otherwise the elseaction
  (if there is one) will be peformed.

  examples:
    #if {$myhpvar < 100} {#showme PANIC!}
    #if {$myhpvar < 100 and $myspvar < 100} {#showme PANIC!}
    #if {'$name' == 'Joe'} {#showme That joe is a jerk.}

  When you're comparing variable values with other strings, make sure 
  to put them in quotes becuase variable expansion happens before
  the if command is evaluated.

  examples:
    WRONG: #if {$name == Joe} {#showme Joe is a jerk.}
    RIGHT: #if {'$name' == 'Joe'} {#showme Joe is a jerk.}

  category: commands
  """
  # original if_cmd code contributed by Sebastian John

  expr = args["expr"]
  action = args["action"]
  elseaction = args["elseaction"]

  try:
    if eval(expr):
      exported.lyntin_command(action, 1, ses)
    elif elseaction:
      exported.lyntin_command(elseaction, 1, ses)
  except SyntaxError:
    exported.write_error("if: invalid syntax / syntax error.", ses)
  except Exception as e:
    exported.write_error("if: exception: %s" % e, ses)

commands_dict["if"] = (if_cmd, "expr action elseaction=")


def info_cmd(ses, args, input):
  """
  Prints all the information about the active session: 
  actions, aliases, gags, highlights, variables, ticker, verbose, 
  speedwalking, and other various things.

  category: commands
  """
  data = exported.myengine.getStatus(ses)
  data = "\n".join(data)
  exported.write_message(data, ses)

commands_dict["info"] = (info_cmd, "")


def killall_cmd(ses, args, input):
  """
  Clears all sessions of session oriented stuff: aliases,
  substitutions, gags, variables, so on so forth.

  category: commands
  """
  for mem in exported.get_active_sessions():
    mem.clear()
    exported.write_message("killall: session %s cleared." % mem.getName())

commands_dict["^killall"] = (killall_cmd, "")


def loop_cmd(ses, args, input):
  """
  Executes a given command replacing %0 in the command with
  the range of numbers specified in <from> and <to>.

  example:
    #loop {1,4} {reclaim %0.corpse}

  will execute:
    reclaim 1.corpse
    reclaim 2.corpse
    reclaim 3.corpse
    reclaim 4.corpse

  Additionally, it can iterate over a comma-separated string of items:

  example:
    #loop {joe,harry,fred,pete} {say hi, %0.} range=no

  will execute:

    say hi, joe.
    say hi, harry.
    say hi, fred.
    say hi, pete.

  A better way to execute a command a number of times without regard
  to an index, would be:

    #4 {reclaim corpse}

  which will send "reclaim corpse" to the mud 5 times.

  category: commands
  """
  loop = args["fromto"]
  command = args["comm"]
  r = args["range"]

  # split it into parts
  loopitems = loop.split(',')

  if r:
    if len(loopitems) != 2:
      exported.write_error("syntax: #loop <from,to> <command>", ses)
      return

    # remove trailing and leading whitespace and convert to ints
    # so we can use them in a range function
    ifrom = int(loopitems[0].strip())
    ito = int(loopitems[1].strip())

    # we need to handle backwards and forwards using the step
    # and need to adjust ito so the range is correctly bounded.
    if ifrom > ito:
      step = -1
    else:
      step = 1

    if ito > 0:
      ito = ito + step
    else:
      ito = ito - step

    loopitems = list(range(ifrom, ito, step))
    loopitems = [repr(m) for m in loopitems]
    # for i in range(ifrom, ito, step):
    #   loopcommand = command.replace("%0", repr(i))
    #   exported.lyntin_command(loopcommand, internal=1, session=ses)

  for mem in loopitems:
    mem = mem.strip()
    loopcommand = command.replace("%0", mem)
    exported.lyntin_command(loopcommand, internal=1, session=ses)

commands_dict["loop"] = (loop_cmd, "fromto comm range:boolean=true")


def math_cmd(ses, args, input):
  """
  Implements the #math command which allows you to manipulate
  variables above and beyond setting them.

  examples:
    #math {hps} {$hps + 5}

  category: commands
  """
  var = args["var"]
  ops = args["operation"]
  quiet = args["quiet"]

  try:
    rvalue = eval(ops)
    varman = exported.get_manager("variable")
    if varman:
      varman.addVariable(ses,var, str(rvalue))
    if not quiet:
      exported.write_message("math: %s = %s = %s." % (var, ops, str(rvalue)), ses)
  except Exception as e:
    exported.write_error("math: exception: %s\n%s" % (ops, e), ses)

commands_dict["math"] = (math_cmd, "var operation quiet:boolean=false")


def nop_cmd(ses, args, input):
  """
  nop stands for "no operation".  So anything after a #nop
  and before a ; (unless it's escaped) will be ignored.

  This was quite possibly the easiest command to program ever.

  category: commands
  """
  return

commands_dict["nop"] = (nop_cmd, "input=", "limitparsing=0")


def read_cmd(ses, args, input):
  """
  Reads in a file running each line as a Lyntin command.  This is the
  opposite of #write which allows you to save session settings and
  restore them using #read.

  You can also read in via the commandline when you start Lyntin:

    lyntin --read 3k

  And read can handle HTTP urls:

    lyntin --read http://lyntin.sourceforge.net/lyntinrc

    #read http://lyntin.sourceforge.net/lyntinrc

  Note: the first non-whitespace char is used to set the Lyntin
  command character.  If you use non Lyntin commands in your file,
  make sure the first one is a command char.  If not, use #nop .
  It will skip blank lines.

  If you don't specify a directory, Lyntin will look for the file
  in your datadir.

  category: commands
  """
  filename = args.get("filename")

  if os.sep not in filename and not filename.startswith("http://"):

    if config.options["datadir"] is None:
      raise ValueError("Error: config.options['datadir'] is not set!")

    if filename is None:
      raise ValueError("Error: filename is None!")
    filename = config.options["datadir"] + filename
    #TODO i have a file, just nor parsing it properly!
    #filename = os.path.join(config.options["datadir"], filename)

  if filename.startswith("~"):
    filename = os.path.expanduser(filename)

  try:
    # http reading contributed by Sebastian John
    if filename.startswith("http://"):
      contents = utils.http_get(filename).split("\n")
    else:
      f = open(filename, "r")
      contents = f.readlines()
      f.close()
  except Exception as e:
    exported.write_error("read: file %s cannot be opened.\n%s" % (filename, e), ses)
    return

  contents = [m for m in contents if len(m.strip()) > 0]

  if len(contents) == 0:
    exported.write_message("read: %s had no data." % filename, ses)
    return

  c = exported.get_config("commandchar")
  if not contents[0].startswith(c):
    exported.lyntin_command("%sconfig commandchar %s" % (c, contents[0][0]), internal=1, session=ses)

  command = ""
  continued = 0
  # FIXME - should this be a config setting?
  esc = "\\"

  for mem in contents:
    mem = mem.strip()
    if len(mem) > 0:
      # handle multi-line commands
      if mem.endswith(esc):
        mem = mem.rstrip(esc)
        continued = 1
      else:
        continued = 0

    command = command + mem
    if not continued:
      exported.lyntin_command(command, internal=1, session=ses)
      command = ""

  exported.write_message("read: file %s read." % filename, ses)

commands_dict["read"] = (read_cmd, "filename")
def session_cmd(ses, args, input):
  """
  This command creates a connection to a specific mud. When you create
  a session, that session becomes the active Lyntin session.

  To create a session to 3k.org named "3k":

  #session 3k www.3k.org 5000

  To create a session and initialize it with commands from a specific
  file:

  #session 3k www.3k.org 5000 /home/david/3k/3k.lyntin

  Then to create another session to another mud:

  #session eto gytje.pvv.unit.no 4000

  Then if 3k was your active session, you could do things on the eto
  session by prepending your command with "#eto ":

  #eto say hello

  or switch to the eto session by typing just "#eto".

  category: commands
  """
  name = args["sessionname"]
  host = args["host"]
  port = args["port"]
  filename = args["filename"]

  if not name and not host and (not port or port == -1):
    data = "Sessions available:\n"
    for mem in exported.get_active_sessions():
      data += f"   {mem.getName()}: {mem._socket!r}\n"

    exported.write_message(data[:-1])
    return

  if not name or not host or port == -1:
    exported.write_error("syntax: #session <sesname> <host> <port>")
    return

  if name.isdigit():
    exported.write_error("session: session name cannot be all numbers.")
    return

  e = exported.myengine
  ses = e.getSession(name)

  preexistingsession = ses is not None

  if preexistingsession and ses.isConnected():
    exported.write_error("session: session of that name already exists.")
    return

  try:
    # create and register a session for this connection....
    if ses is None:
      ses = e.createSession(name)

    sock = net.SocketCommunicator(e, ses, host, port)
    ses.setSocketCommunicator(sock)

    ses._host = host
    ses._port = port

    e.changeSession(name)

    # connect to the mud...
    # this might take a while--we block here until this is done.
    sock.connect(host, port, name)

    # start the network thread
    e.startthread("network", sock.run)

  except Exception as e:
    exported.write_traceback(f"session: had problems creating the session. {str(e)}")
    ses.setSocketCommunicator(None)

    if not preexistingsession:
      try:
        e.unregisterSession(ses)
      except Exception:
        pass

      try:
        e.closeSession(name)
      except Exception:
        pass

  # populate the session using the specified file
  if filename:
    read_cmd(ses, args, '')


commands_dict["session"] = (session_cmd, "sessionname= host= port:int=-1 filename=")


def showme_cmd(ses, args, input):
  """
  Will display {text} on your screen.  Doesn't get sent to the mud--
  just your screen.

  examples:
    #action {^%0 annihilates you!} {#showme {EJECT! EJECT! EJECT!}}

  category: commands
  """
  input = args["input"]
  if not input:
    exported.write_error("syntax: requires a message.", ses)
    return

  input = input.replace("\\;", ";")
  input = input.replace("\\$", "$")
  input = input.replace("\\%", "%")

  exported.write_message(input, ses)
     
commands_dict["showme"] = (showme_cmd, "input=", "limitparsing=0")

def wshowme_cmd(ses, args, input):
  """
  Writes the text into the named window, if the current ui supports named window
s.
  If named windows are unsupported, writes the text into the main window.

  examples:
    #action {^%0 annihilates you!} {#wshowme Alert {EJECT! EJECT! EJECT!}}

  category: commands
  """
  exported.write_message(args["text"], ses, window=args["window"])

commands_dict['wshowme'] = ( wshowme_cmd, "window= text=" )


def textin_cmd(ses, args, input):
  """
  Takes the contents of the file and outputs it directly to the mud
  without processing it (like #read does).

  If you don't specify a directory, Lyntin will look for the file in
  the datadir.

  category: commands
  """
  if (ses.getName() == "common"):
    exported.write_error("textin cannot be applied to common session.", ses)
    return

  filename = args["file"]

  if os.sep not in filename:
    #filename = config.options["datadir"] + filename
    filename = os.path.join(config.options["datadir"], filename)
   
  try:
    f = open(filename, "r")
    contents = f.readlines()
    f.close()
    for mem in contents:
      mem = utils.chomp(mem)
      ses.getSocketCommunicator().write(mem + "\n")
    exported.write_message("textin: file %s read and sent to mud." % filename, ses)

  except OSError:
    exported.write_error("textin: file %s is not readable." % filename, ses)
  except Exception as e:
    exported.write_error("textin: exception thrown %s." % e, ses)

commands_dict["textin"] = (textin_cmd, "file")


def version_cmd(ses, args, input):
  """
  Displays the version number, contact information, and web-site for
  Lyntin.

  category: commands
  """
  exported.write_message(constants.VERSION)

commands_dict["version"] = (version_cmd, "")


def write_cmd(ses, args, input):
  """
  Writes all aliases, actions, gags, etc to the file specified.
  You can then use the #read command to read this file in and
  restore your session settings.

  The quiet argument lets you specify whether you want command data
  to be written to the file so that when you read it back in with #read,
  the commands are executed quietly.

  If you don't specify a directory, it will be written to your datadir.

  Note: Windows users should either use two \\'s or use / to separate
  directory names.

  category: commands
  """
  filename = args["file"]
  quiet = args["quiet"]

  f = None

  c = exported.myengine.getConfigManager().get("commandchar")

  if os.sep not in filename:
    filename = config.options["datadir"] + filename

  data = exported.get_write_data(ses, quiet)

  if data:
    try:
      f = open(filename, "w")
      f.write(c + ("\n" + c).join(data))
      f.close()
      exported.write_message("write: file %s has been written for session %s." % 
                             (filename, ses.getName()), ses)
    except Exception:
      exported.write_traceback("write: error writing to file %s." % filename, ses)
      try:
        f.close()
      except:
        pass
    return

  exported.write_message("write: no data to write.")

commands_dict["write"] = (write_cmd, "file quiet:boolean=false")


def zap_cmd(ses, args, input):
  """
  This disconnects from the mud and closes the session.  If no
  session is specified, it will close the current session.

  category: commands
  """
  sesname = args["session"]
  if sesname:
    ses = exported.myengine.getSession(sesname)
    if ses is None:
      exported.write_error("zap: session %s does not exist." % sesname)
      return

  if exported.myengine.closeSession(ses):
    exported.write_message("zap: session %s zapped!" % ses.getName())
  else:
    exported.write_message("zap: session %s cannot be zapped!" % ses.getName())

commands_dict["zap"] = (zap_cmd, "session=")


def load():
  """ Initializes the module by binding all the commands."""
  modutils.load_commands(commands_dict)


def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  modutils.unload_commands(list(commands_dict.keys()))

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: history.py,v 1.4 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
The HistoryManager keeps track of the last 1000 lines of user input 
X{history}. The HistoryManager is a singleton and it's on an engine 
scoping thus we don't keep track of history per session.
"""
from lyntin import manager

class HistoryManager(manager.Manager):
  """
  Manages user data history.

  This manages the user input history by storing the commands the
  user entered.  This module also handles manipulating that history
  letting the user to recall and edit those commands to fix mistakes
  they may have typed.
  """
  def __init__(self, e):
    self._history = [""]
    self._config = e.getManager("config")
    self._engine = e

  def getHistoryItem(self, userinput):
    """
    This retrieves the item (if it exists) and performs the 
    substitutions (if we need to).

    @param userinput: what the user typed--we'll use this to figure
        out which item they're referring to and whether to apply a 
        substitution
    @type  userinput: string

    @returns: None if we didn't discover anything or the command 
        string at the history index
    """
    tokens = userinput.split(" ", 1)

    # grab the first (and possibly only) token and remove the !
    index = tokens[0][1:]

    # if it's very short, we're looking at the last thing typed
    # (prior to this thing they typed)
    if len(index) == 0:
      returninput = self._history[0]
    else:
      try:
        returninput = self._history[int(index)]
      except:
        for h in self._history:
          if h[0:len(index)] == index:
            return h
        return None

    # check to see if they want to do a substitution
    if len(tokens) > 1:
      i = tokens[1].find("=")
      if i != -1:
        returninput = returninput.replace(tokens[1][:i], tokens[1][i+1:])

    return returninput

  def getHistory(self, count):
    """
    Returns everything in the history buffer as a list of strings

    @return: everything in the history buffer
    @rtype: list of strings
    """
    return self._history[:count]

  def recordHistory(self, input):
    """
    Records an item in the history (which is a queue).

    @param input: the line to record
    @type  input: string
    """
    if not input:
      return

    if input == self._history[0] and not self._config.get("repeathistory"):
      return

    self._history.insert(0, input)
    if len(self._history) > 1000:
      del self._history[-1]

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: engine.py,v 1.36 2007/10/02 22:06:32 willhelm Exp $
#########################################################################
"""
This holds the X{engine} which both contains most of the other objects
that do work in Lyntin as well as encapsulates the event queue, event
handling methods, and some of the other singleton managers such as
the HelpManager and the CommandManager.

Engine also holds hooks to the various event types.  Events will call
all appropriate hooks allowing you to add functionality via the modules
interface without changing the Lyntin internals.

The engine also holds a list of registered threads.  This helps in
diagnostics.  Use the methods in exported to handle spinning off
threads.

The Engine class is a singleton and the reference to it is stored in
"engine.Engine.instance".  However, you should use the exported module
to access the engine using the "get_engine()" function.


X{startup_hook}::

   The startup hook is called after Lyntin has bootstrapped itself
   enough to allow everything to initialize itself.

   Arg mapping: {}


X{shutdown_hook}::

   When Lyntin is shutting down, this hook is called.  It's possible
   Lyntin might be in a state of disarray at this point, so it's not
   clear what is and what is not available.

   Arg mapping: {}


X{timer_hook}::

   The timer hook spams all registered functions every second.  This
   is how the scheduler works.

   Arg mapping: { "tick": int }

   tick - the current tick (starts at 0)


X{from_user_hook}::

   All input typed in by the user (as well as other things that
   eventually go through the handleUserData method) get passed
   through this hook (unless it's specified as internal).  All
   registered functions get to see the raw user data this way.

   Arg mapping: {"data": string}

   data - the user data passed into handleUserData


X{to_user_hook}::

   This hook is for data to be displayed to the user.  The UI listens
   on this hook as do logger functions.

   NOTE: Functions registered with this hook should NEVER call 
   exported.write* functions.  That will result in an infinite loop and 
   then Lyntin will either hang or die.  If you want to spit out output
   use an OutputEvent.

   Arg mapping: { "message": string }

   message - The message to be displayed.  This can be either a string
             of a ui.base.Message object.


X{error_occurred_hook}::

   Every time an event kicks up an unhandled error, we add one to
   our error count and also spam this hook with the current
   number of errors.

   Arg mapping: { "count": int }

   count - the current number of errors


X{too_many_errors_hook}::

   When we hit the maximum number of errors, this hook gets spammed and
   then Lyntin shuts down.

   Arg mapping: {}


X{session_change_hook}::

   This hook gets spammed whenever the current session changes,
   after changing.

   Arg mapping: { "new": session.Session, "previous": session.Session }

   new - the session that is being changed to

   previous - the session that was previously the current session
"""
import queue
import threading
import sys
import traceback
import os.path

from threading import Thread

from lyntin import config, session, utils, event, exported, helpmanager, history, commandmanager, constants


class Engine:
  """
  This is the engine class.  There should be only one engine.
  """
  instance = None

  def __init__(self):
    """ Initializes the engine."""

    # this is the event queue that holds all the events in
    # the system.
    self._event_queue = queue.Queue()

    # this is a lock for writing stuff to the ui--makes sure
    # we're not hosing things by having multiple things write
    # to the ui simultaneously....  ick.
    self._ui_lock = threading.Lock()

    # this is the master shutdown flag for the event queue
    # handling.
    self._shutdownflag = 0

    # Lyntin counts the total number of errors it's encountered.
    # This enables us to shut ourselves down if we encounter too
    # many which indicates a "bigger problem".
    self._errorcount = 0

    # listeners exist at an engine level.  if you sign up for
    # an input hook, you get the input hook for ALL sessions.
    # this might change at some point....  we'll see.
    self._listeners = {}

    self._managers = {}

    # the help manager manages all the help content in a hierarchical
    # structure.
    self._managers["help"] = helpmanager.HelpManager(self)

    # our config manager
    self._managers["config"] = config.ConfigManager(self)

    # our history manager
    self._managers["history"] = history.HistoryManager(self)

    # our command manager
    self._managers["command"] = commandmanager.CommandManager(self)

    # there is only one ui in the system.
    self._ui = None

    # current tick count
    self._current_tick = 0

    # list of registered threads
    self._threads = []

    # counts the total number of events processed--for diagnostics
    self._num_events_processed = 0

    # holds all the sessions
    self._sessions = {}

    # the current session.  points to a Session object.
    self._current_session = None

    # map of hook name -> utils.PriorityQueue objects
    self._hooks = {}

    # we register ourselves with the shutdown hook
    self.hookRegister("shutdown_hook", self.shutdown)

    commonsession = session.Session(self)
    commonsession.setName("common")

    # this creates a "common" entry in all the managers that manage
    # session scoped data--the session base is None
    # for mem in self._managers.values():
    #   mem.addSession(commonsession, None)

    self._sessions["common"] = commonsession
    self._current_session = commonsession

    self.hookRegister("user_filter_hook", self._managers["command"].filter, 100)

  def _setupConfiguration(self):
    """
    Goes through and sets up all the engine-specific configuration
    pieces.
    """
    c = self._managers["config"]

    # this one doesn't seem to do anything
    # c.add("variablechar", config.CharConfig("variablechar", "$", 0, "denotes variables"))

    cops = config.options

    c.add("repeathistory", config.BoolConfig("repeathistory", 
          utils.convert_boolean(cops.get("repeathistory", 1)), 0, 
          "Whether (yes) or not (no) we record repeated user input in the " +
          "history buffer.  For example, if you type \"north\" and then " +
          "\"north\" again, if repeathistory is on, we record both.  " +
          "Otherwise we would only record the first one."))

    c.add("commandchar", config.CharConfig("commandchar", 
          config.options.get("commandchar", "#"), 0, 
          "The character used to denote a command."))

    c.add("debugmode", config.BoolConfig("debugmode", 
          utils.convert_boolean(cops.get("debugmode", 0)), 0, 
          "Debug mode helps you to figure out how your commands are being " +
          "evaluated."))

    c.add("promptdetection", config.BoolConfig("promptdetection", 
          utils.convert_boolean(cops.get("promptdetection", 0)), 0, 
          "Prompt detection is done in net.py when mud data comes in.  " +
          "This toggles whether we detect prompts or not.  This won't help " +
          "you unless you have a plugin which requires it."))

    c.add("ansicolor", config.BoolConfig("ansicolor", 
          utils.convert_boolean(cops.get("ansicolor", 1)), 1,
          "Allows you to enable or disable ansi color handling."))

    c.add("mudecho", config.BoolConfig("mudecho", 
          utils.convert_boolean(cops.get("mudecho", 1)), 0,
          "Whether (1) or not (0) we're echoing user input to the ui."))

    c.add("datadir", config.StringConfig("datadir",
          config.options["datadir"], 0,
          "Default directory to find config files etc."))           

    c.add("splitchar", config.CharConfig("splitchar",
          config.options.get("splitchar", ";"), 0,
          "The character used to split commands in a single input."))

    self._sessions["common"].setupCommonSession()

  ### ------------------------------------------
  ### hook stuff
  ### ------------------------------------------
  def checkHooks(self):
    """
    Goes through all the hooks and returns a list of strings of
    basic information about them.

    @returns: information about the hooks
    @rtype: list of strings
    """
    data = []
    for mem in list(self._hooks.keys()):
      data.append("   %s - %d registered functions" % (mem, self._hooks[mem].count()))

    return data

  def getHook(self, hookname, newhook=1):
    """
    Retrieves the hook in question.  If the hook doesn't 
    exist and newhook==1, then we'll create a new hook.
    Otherwise, we'll return None.

    @param hookname: the name of the hook to retrieve
    @type  hookname: string

    @returns: the hook by name
    @rtype: utils.PriorityQueue
    """
    if hookname in self._hooks:
      return self._hooks[hookname]

    if newhook==1:
      self._hooks[hookname] = utils.PriorityQueue()
      return self._hooks[hookname]

    return None

  def hookRegister(self, hookname, func, place=constants.LAST):
    """
    Registers a function with a hook.

    @param hookname: the name of the hook
    @type  hookname: string

    @param func: the function to register with the hook
    @type  func: function

    @param place: the function will get this place in the call
        order.  functions with the same place specified will get
        arbitrary ordering.  defaults to constants.LAST.
    @type  place: int
    """
    hook = self.getHook(hookname)
    if place is None:
      hook.add(func)
    else:
      hook.add(func, place)
 
  ### ------------------------------------------
  ### thread stuff
  ### ------------------------------------------

  def startthread(self, name, func):
    """
    Starts a thread through the Thread Manager.

    @param name: the name of the thread to start
    @type  name: string

    @param func: the function to run in the thread
    @type  func: function

    @return:     Thread instance
    @rtype:      threading.Thread
    """
    # clean up the list of threads that we maintain first
    self._threadCleanup()

    # create and initialize the new thread and stick it in our list
    t = Thread(None, func)
    t.setDaemon(1)
    t.setName(name)
    t.start()
    self._threads.append(t)
    return t

  def checkthreads(self):
    """
    Calls the Thread Manager checkthreads method which goes
    through and checks the status of all the threads registered
    with the Thread Manager.

    @return: one string for each thread indicating its status
    @rtype: list of strings
    """
    data = []
    alive = { 0: "not alive", 1: "alive" }

    for mem in self._threads:
      data.append("   %s - %s" % (mem.getName(), alive[mem.is_alive()]))

    return data

  def _threadCleanup(self):
    """
    Removes threads which have ended.
    """
    removeme = []
    for i in range(len(self._threads)):
      if self._threads[i].is_alive() == 0:
        removeme.append(self._threads[i])

    for mem in removeme:
      self._threads.remove(mem)


  ### ------------------------------------------
  ### timer thread
  ### ------------------------------------------

  def runtimer(self):
    """
    This timer thread sleeps for a second, then calls everything
    in the queue with the current tick.
    """
    from time import time
    
    self._current_tick = 0
    wakeup_time = current_time = time()

    import threading
    ev = threading.Event()
    exported.hook_register("shutdown_hook", lambda *_: ev.set())

    while not ev.isSet():
      try:
        tout = wakeup_time - current_time
        if tout > 0:
          ev.wait(timeout = tout)
          current_time = time()
        elif tout < -10:
          # we are late too much; drop 10 ticks
          wakeup_time += 10
          continue
        event.SpamEvent(hookname="timer_hook", 
                        argmap={"tick": self._current_tick}
                       ).enqueue()
        self._current_tick += 1
        wakeup_time += 1
      except KeyboardInterrupt:
        return
      except SystemExit:
        return
      except:
        exported.write_traceback("ticker: ticker hiccupped.")
    

  ### ------------------------------------------
  ### input/output stuff
  ### ------------------------------------------

  def handleUserData(self, input, internal=0, session=None):
    """
    This handles input lines from the user in a session-less context.
    The engine.handleUserData deals with global stuff and then
    passes the modified input to the session for session-oriented
    handling. The session can call this method again with
    expanded input--this method is considered recursive.

    internal tells whether to spam the input hook and
    things of that nature.

    @param input: the data from the user
    @type  input: string

    @param internal: whether this should be executed internally or not.
        0 if we should spam the input hook and record
        the input to the historymanager; 1 if we shouldn't
    @type  internal: boolean

    @param session: the session scoping to execute this user input in
    @type  session: session.Session instance

    @return: the commands that were actually executed (may not be
        exactly what the user typed--this is for the history manager)
    @rtype: string
    """
    if self._managers["config"].get("debugmode") == 1:
      exported.write_message("evaluating: %s" % input)

    inputlist = utils.split_commands(
      self._managers["config"].get("splitchar"), input
    )
    if session is None:  # Use `is None` for comparison
      session = self._current_session

    historyitems = []
    commandchar = self._managers["config"].get("commandchar")
    for mem in inputlist:
      # mem = mem.strip()

      if len(mem) == 0:
        mem = commandchar + "cr"

      # if it's not internal we spam the hook with the raw input
      if internal == 0:
        exported.hook_spam("from_user_hook", {"data": mem})

      if mem.startswith("!"):
        memhistory = self.getManager("history").getHistoryItem(mem)
        if memhistory is not None:  # Use `is not None` for comparison
          self.handleUserData(memhistory, 1, session)
          historyitems.append(memhistory)
          continue

      # if it starts with a # it's a loop, session, or command
      if len(mem) > 0 and mem.startswith(commandchar):

        # pull off the first token without the commandchar
        ses = mem.split(" ", 1)[0][1:]

        # is it a loop (aka repeating command)?
        if ses.isdigit():
          num = int(ses)
          if mem.find(" ") != -1:
            command = mem.split(" ", 1)[1]
            command = utils.strip_braces(command)
            if num > 0:
              for i in range(num):
                loopcommand = self.handleUserData(command, 1, session)
              historyitems.append(commandchar + ses + " {" + loopcommand + "}")
          continue

        # is it a session?
        if ses in self._sessions:
          input = mem.split(" ", 1)
          if len(input) < 2:
            self.set_current_session(self._sessions[ses])
          else:
            self.handleUserData(
              input[1], internal=1, session=self._sessions[ses]
            )
          historyitems.append(mem)
          continue

        # is it "all" sessions?
        if ses == "all":
          newinput = mem.split(" ", 1)
          if len(newinput) > 1:
            newinput = newinput[1]
          else:
            newinput = commandchar + "cr"

          for sessionname in self._sessions:  # Iterate directly over the dict
            if sessionname != "common":
              self._sessions[sessionname].handleUserData(newinput, internal)
          historyitems.append(mem)
          continue

      # if we get here then it is not a valid !-expression and it's going
      # to the default session
      historyitems.append(mem)

      # no command char, so we pass it on to the session.handleUserData
      # to do session-oriented things
      session.handleUserData(mem, internal)

    # we don't record internal stuff or input that isn't supposed
    # to be echoed
    executed = ";".join(historyitems)
    if internal == 0 and self.getConfigManager().get("mudecho") == 1:
      self.getManager("history").recordHistory(executed)

    return executed

  def handleMudData(self, session, text):
    """
    Handle input coming from the mud.  We toss this to the 
    current session to deal with.

    @param session: the session this mud data applies to
    @type  session: session.Session instance

    @param text: the text coming from the mud
    @type  text: string
    """
    if session:
      session.handleMudData(text)
    else:
      exported.write_message("Unhandled data:\n%s" % text)


  ### ------------------------------------------
  ### session stuff
  ### ------------------------------------------

  def createSession(self, name):
    """
    Creates a new session by copying the common session
    and registers the new session with the engine.

    @param name: the name of the session
    @type  name: string

    @return: the new session
    @rtype: session.Session instance
    """
    ses = session.Session(self)
    ses.setName(name)
    self.registerSession(ses, name)
    return ses

  def registerSession(self, session, name):
    """
    Registers a session with the engine.

    @param session: the session to register
    @type  session: session.Session instance

    @param name: the name of the session
    @type  name: string

    @raises ValueError: if the session has a non-unique name
    """
    if name in self._sessions:
      raise ValueError("Session of that name already exists.")

    commonsession = self.getSession("common")
    for mem in list(self._managers.values()):
      mem.addSession(session, commonsession)

    self._sessions[name] = session

  def unregisterSession(self, ses):
    """
    Unregisters a session from the engine.

    @param ses: the session to unregister
    @type  ses: session.Session instance
    """
    if ses.getName() not in self._sessions:
      raise ValueError("No session of that name.")

    if ses == self._current_session:
      self.changeSession()

    for mem in list(self._managers.values()):
      try:
        mem.removeSession(ses)
      except Exception as e:
        exported.write_error("Exception with removing session %s." % e)

    del self._sessions[ses.getName()]

  def getSessions(self):
    """
    Returns a list of session names.

    @return: all the session names
    @rtype: list of strings
    """
    return list(self._sessions.keys())

  def getSession(self, name):
    """
    Returns a named session.

    @param name: the name of the session to retrieve
    @type  name: string

    @return: the session of that name or None
    @rtype: session.Session or None
    """
    if name in self._sessions:
      return self._sessions[name]
    else:
      return None

  def changeSession(self, name=''):
    """
    Changes the current session to another named session.

    If no name is provided, switch to the next available non-common session if possible.

    @param name: The name of the session to switch to
    @type name: str
    """
    try:
      if not self._sessions:
        # Handle the case where no sessions are available
        exported.write_error("No available sessions to switch to.")
        return

      if name == '':
        # Get all session names as a list
        session_names = list(self._sessions.keys())

        # Get the current session name, if available
        current_session_name = (
          self._current_session.getName() if self._current_session else None
        )

        # Exclude the current session name from the list
        if current_session_name in session_names:
          session_names.remove(current_session_name)

        # Ensure "common" is the last option
        if "common" in session_names:
          session_names.remove("common")
          session_names.append("common")

        if session_names:
          # Switch to the first session in the list
          self.set_current_session(self._sessions[session_names[0]])
        else:
          # No valid sessions to switch to
          exported.write_error("No other sessions to switch to.")
      elif name in self._sessions:
        # Switch to the specified session by name
        self.set_current_session(self._sessions[name])
      else:
        # Handle invalid session names
        exported.write_error(f"No session with the name '{name}' exists.")
    except Exception as e:
      # Log unexpected errors
      exported.write_error(f"Error while changing session: {e}")

  def set_current_session(self, newsession):
    """
    Changes the current session to another session.

    @param newsession: the session to change to
    @type  newsession: session.Session instance
    """
    previous_session = self._current_session
    self._current_session = newsession
    exported.hook_spam("session_change_hook", {"new": newsession,
                                      "previous": previous_session})
    exported.write_message("%s now current session." % newsession.getName())

  def writeSession(self, message):
    """
    Writes a message to the network socket.  The message should 
    be a string.  Otherwise, it's unhealthy.

    @param message: the text to write to the mud
    @type  message: string
    """
    self._current_session.write(message)

  def closeSession(self, ses=None):
    """
    Closes down a session.

    @param ses: the name of the session to close
    @type  ses: string

    @return: 1 if successful; 0 if not
    @rtype: boolean
    """
    if ses == None:
      ses = self._current_session

    if ses.getName() == "common":
      exported.write_error("Can't close the common session.")
      return 0
         
    ses.shutdown((1,))
    self.unregisterSession(ses)
    exported.hook_unregister("shutdown_hook", ses.shutdown)
    return 1


  ### ------------------------------------------
  ### event-handling/engine stuff
  ### ------------------------------------------

  def _enqueue(self, event):
    """
    Adds an event to the queue.

    @param event: the new event to enqueue
    @type  event: event.Event
    """
    self._event_queue.put(event)

  def runengine(self):
    """
    This gets kicked off in a thread and just keep going through
    events until it detects a shutdown.
    """
    while not self._shutdownflag:
      try:
        # blocks on the event queue
        e = self._event_queue.get()
        e.execute()
      except KeyboardInterrupt:
        return
      except SystemExit:
        return
      except:
        self.tallyError()
        exported.write_traceback("engine: unhandled error in engine.")
      self._num_events_processed += 1

        
  def tallyError(self):
    """
    Adds one to the error count.  If we see more than 20 errors, we shutdown.
    """
    self._errorcount = self._errorcount + 1
    exported.write_error("WARNING: Unhandled error encountered (%d out of %d)." 
                         % (self._errorcount, 20))
    exported.hook_spam("error_occurred_hook", {"count": self._errorcount})

    if self._errorcount > 20:
      exported.hook_spam("too_many_errors_hook", {})
      exported.write_error("Error count exceeded--shutting down.")
      sys.exit(0)


  def shutdown(self, args):
    """ Sets the shutdown status for the engine."""
    self._shutdownflag = 1

  def getDiagnostics(self):
    """
    Returns some basic diagnostic information in the form of a string.
    This allows a user to monitor how Lyntin is doing in terms
    of events and other such erata.

    @return: the complete the complete diagnostic data for our little happy
        mud client
    @rtype: string
    """
    data = []
    data.append("   events processed: %d" % self._num_events_processed)
    data.append("   queue size: %d" % self._event_queue.qsize())
    data.append("   ui: %s" % repr(self._ui))
    data.append("   ansicolor: %d" % self.getConfigManager().get("ansicolor"))
    data.append("   ticks: %d" % self._current_tick)
    data.append("   errors: %d" % self._errorcount)

    # print info from each session
    data.append("Sessions:")
    data.append("   total sessions: %d" % len(self._sessions))
    data.append("   current session: %s" % self._current_session.getName())

    for mem in list(self._sessions.values()):
      # we do some fancy footwork here to make it print nicely
      info = "\n   ".join(self.getStatus(mem))
      data.append('   %s\n' % info)

    return "\n".join(data)


  ### ------------------------------------------
  ### user interface stuff
  ### ------------------------------------------

  def setUI(self, newui):
    """
    Sets the ui.

    @param newui: the new ui to set
    @type  newui: ui.base.BaseUI subclass
    """
    self._ui = newui

  def getUI(self):
    """
    Returns the ui.

    @return: the ui
    @rtype: ui.base.BaseUI subclass
    """
    return self._ui

  def writeUI(self, text):
    """
    Writes a message to the ui.

    This method uses a lock so that multiple threads can write
    to the ui without intersecting and crashing the python process.

    Theoretically you should use the exported module to write
    things to the ui--it calls this method.

    @param text: the message to write to the ui
    @type  text: string or ui.base.Message
    """
    self._ui_lock.acquire(1)
    try:
      exported.hook_spam("to_user_hook", {"message": text})
    finally:
      self._ui_lock.release()

  def writePrompt(self):
    """ Tells the ui to print a prompt."""
    if self._ui:
      self._ui.prompt()

  def flushUI(self):
    """ Tells the ui to flush its output."""
    self._ui.flush()

  
  ### ------------------------------------------------
  ### config functions
  ### ------------------------------------------------
  def getConfigManager(self):
    """
    Returns the config manager.
    """
    return self._managers["config"]

  ### ------------------------------------------------
  ### Manager functions
  ### ------------------------------------------------

  def addManager(self, name, manager):
    """
    Adds a manager to our list.

    @param name: the name of the manager to add
    @type  name: string

    @param manager: the manager instance to add
    @type  manager: manager.Manager subclass
    """
    self._managers[name] = manager

  def removeManager(self, name):
    """
    Removes a manager from our list.

    @param name: the name of the manager to remove
    @type  name: string

    @return: 0 if nothing happened, 1 if the manager was removed
    @rtype: boolean
    """
    if name in self._managers:
      del self._managers[name]
      return 1
    return 0

  def getManager(self, name):
    """
    Retrieves a manager by name.

    @param name: the name of the manager to retrieve
    @type  name: string

    @return: the manager instance or None
    @rtype: manager.Manager subclass or None
    """
    if name in self._managers:
      return self._managers[name]
    return None

  ### ------------------------------------------------
  ### Status stuff
  ### ------------------------------------------------
  def getStatus(self, ses):
    """
    Gets the status for a specific session.

    @param ses: the session to get status for
    @type  ses: session.Session

    @return: the status of the session
    @rtype: list of strings
    """
    # call session.getStatus() and get status from it too
    data = ses.getStatus()

    # loop through our managers and get status from them
    managerkeys = sorted(self._managers.keys())

    for mem in managerkeys:
      temp = self.getManager(mem).getStatus(ses)
      if temp:
        data.append(f"   {mem}: {temp}")  # Use f-strings for better readability

    # return the list of elements
    return data


def main(defaultoptions={}):
  """
  This parses the command line arguments and makes sure they're all valid,
  instantiates a ui, does some setup, spins off an engine thread, and
  goes into the ui's mainloop.

  @param defaultoptions: the boot options to use.  we update the 
      config.options dict with these options--this is the easiest
      way to override the ui, moduledir, datadir, et al from a
      Lyntin run script.
  @type  defaultoptions: dict
  """
  try:
    import sys, os, traceback, configparser
    from lyntin import config, event, utils, exported
    from lyntin.ui import base
    import locale

    locale.setlocale(locale.LC_ALL, '')

    config.options.update(defaultoptions)

    # read through options and arguments
    optlist = utils.parse_args(sys.argv[1:])

    for mem in optlist:
      if mem[0] == '--help':
        print(constants.HELPTEXT)
        sys.exit(0)

      elif mem[0] == '--version':
        print(constants.VERSION)
        sys.exit(0)


      elif mem[0] in ["--configuration", "-c"]:
        # ini files OVERRIDE the default options
        # they can provide multiple ini files, but each new
        # ini file will OVERRIDE the contents of the previous ini file
        # where the two files intersect.
        parser = configparser.ConfigParser()

        if mem[0] is not None:
          parser.read([mem[1]])
        else:
          print("Error: -c option requires a filename")
          sys.exit(1)

        #parser.read([mem[1]])

        newoptions = {}
        for s in parser.sections():
          for o in parser.options(s):
            c = parser.get(s, o).split(",")
            if o in newoptions:
              newoptions[o] += c
            else:
              newoptions[o] = c
            
        config.options.update(newoptions)

      else:
        opt = mem[0]
        while opt.startswith("-"):
          opt = opt[1:]

        if len(opt) > 0:
          if opt in config.options:
            if type(config.options[opt]) is list:
              config.options[opt].append(mem[1])
            else:
              config.options[opt] = mem[1]
          else:
            config.options[opt] = [mem[1]]

    for mem in ["datadir", "ui", "commandchar"]:
      if mem in config.options and isinstance(config.options[mem], list):
        config.options[mem] = config.options[mem][0]

    # if they haven't set the datadir via the command line, then
    # we go see if they have a HOME in their environment variables....
    if not config.options["datadir"]:
      if "HOME" in os.environ:
        config.options["datadir"] = os.environ["HOME"]
    config.options["datadir"] = utils.fixdir(config.options["datadir"])

    def on_shutdown():
      """
      This gets called by the Python interpreter atexit.  The reason
      we do shutdown stuff here is we're more likely to catch things
      here than we are to let everything cycle through the 
      ShutdownEvent.  This should probably get fixed up at some point
      in the future.
      """
      sys.stderr.write("goodbye.\n")
      #exported.hook_spam("shutdown_hook", {})
    import atexit
    atexit.register(on_shutdown)

    # instantiate the engine
    Engine.instance = Engine()
    exported.myengine = Engine.instance
    Engine.instance._setupConfiguration()

    # instantiate the ui
    uiinstance = None
    try:
      uiname = str(config.options['ui'])
      modulename = uiname + "ui"
      uiinstance = base.get_ui(modulename)
      if not uiinstance:
        raise ValueError("No ui instance.")
    except Exception as e:
      print((f"Cannot start '{uiname}': {e}"))
      traceback.print_exc()
      sys.exit(0)

    Engine.instance.setUI(uiinstance)

    # do some more silly initialization stuff
    # adds the .lyntinrc file to the readfile list if it exists.
    if config.options["datadir"]:
      lyntinrcfile = config.options["datadir"] + ".lyntinrc"
      if os.path.exists(lyntinrcfile):
        # we want the .lyntinrc file read in first, so then other
        # files can overwrite the contents therein
        config.options['readfile'].insert(0, lyntinrcfile)
  
    # import modules listed in modulesinit
    exported.write_message("Loading Lyntin modules.")
  
    try:
      from lyntin import modules
      modules.load_modules()
    except:
      exported.write_traceback("Modules did not load correctly.")
      sys.exit(1)
  
    # spam the startup hook 
    exported.hook_spam("startup_hook", {})
  
    commandchar = Engine.instance._managers["config"].get("commandchar")

    # handle command files
    for mem in config.options['readfile']:
      exported.write_message("Reading in file " + mem)
      # we have to escape windows os separators because \ has a specific
      # meaning in the argparser
      mem = mem.replace("\\", "\\\\")
      exported.lyntin_command("%sread %s" % (commandchar, mem), internal=1)
  
    # we're done initialization!
    exported.write_message(constants.STARTUPTEXT)
    Engine.instance.writePrompt()

    engine_thread = Engine.instance.startthread("engine", Engine.instance.runengine)
    timer_thread = Engine.instance.startthread("timer", Engine.instance.runtimer)
    try:
      Engine.instance._ui.runui()
    finally:
      sys.stderr.write("Shutting down...")
      event.ShutdownEvent().enqueue()
      engine_thread.join(10)
      timer_thread.join(10)
      
  except:
    import traceback
    traceback.print_exc()
    sys.exit(1)


# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: net.py,v 1.19 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This holds the SocketCommunicator class which handles socket
connections with a mud and polling the connection for data.

X{bell_hook}::

   When the mud sends a bell character, we spam this hook.  Typically
   the ui's will register with this hook and handle the bell however
   they see fit.

   Arg mapping: { "session": Session }

   session - the session that received the bell


X{prompt_hook}::

   We try to do some prompt detection and separate prompts into their
   own events (separate from mud data).

   Arg mapping: { "session": Session, "prompt": string }

   session - the Session that this prompt came from

   prompt - the prompt string


X{connect_hook}::

   This hook gets spammed every time we make a successful connection.

   Arg mapping: { "session": Session, "host": String, "port": int }

   session - the Session object for this connection

   host - the host we connected to

   port - the port number for the host


X{net_read_data_filter}::

   This allows you to filter incoming data before it passes through
   Lyntin.  If you were going to write an MCCP module, this is the
   hook it would register with to decompress incoming mud data.

   Functions that register with this hook should return the dataadj
   if they did nothing or the adjusted dataadj if they transformed it.
   Look at the user_filter_hook examples in the alias and speedwalk
   modules.

   Arg mapping: { "session": Session, "data": String, "dataadj": String }

   session - the Session that this data came from

   data - the original data we got from the mud

   dataadj - the data the previous function in the hook returned


X{net_write_data_filter}::

   This allows you to filter outgoing data after it has passed through
   Lyntin and just before it gets sent out on the socket.  If you were
   going to write an MCCP module, this is the hook it would register
   with to compress outgoing mud data.

   Functions that register with this hook should return the dataadj
   if they did nothing or the adjusted dataadj if they transformed it.
   Look at the user_filter_hook examples in the alias and speedwalk
   modules.

   Arg mapping: { "session": Session, "data": String, "dataadj": String }

   session - the Session that this data is going to

   data - the original data we wanted to send to the mud

   dataadj - the data the previous function in the hook returned


X{net_handle_telnet_option}::

   There are a series of Telnet options that Lyntin doesn't handle.
   So we allow module writers to handle them if they so desire.
   The data argument is the telnet option string.  So if they
   send us 255 251 24, then that's what you're getting.  We handle
   all the buffering of telnet option stuff--so you needn't worry
   about that.

   We send along the IAC DO TERMTYPE kinds of things as well as the
   IAC SB blah blah IAC SE kinds of things.  This is a handler--so
   if you've handled it, raise an exported.StopSpammingException() .

   Arg mapping: { "session": Session, "data": String }

   session - the Session that this telnet option came from

   data - the telnet option itself

"""
import socket, select, re, os

from lyntin import event, config, exported
from lyntin.ui import message

### --------------------------------------------
### CONSTANTS
### --------------------------------------------

# reverse lookup allowing us to see what's going on more easily
# when we're debugging.
# for a list of telnet options: http://www.freesoft.org/CIE/RFC/1700/10.htm
CODES = {255: "IAC",
         254: "DON'T",
         253: "DO",
         252: "WON'T",
         251: "WILL",
         250: "SB",
         249: "GA",
         240: "SE",
         239: "TELOPT_EOR",
         0: "<IS>",
         1: "[<ECHO> or <SEND/MODE>]",
         3: "<SGA>",
         5: "STATUS",
         24: "<TERMTYPE>",
         25: "<EOR>",
         31: "<NegoWindoSize>",
         32: "<TERMSPEED>",
         34: "<Linemode>",
         35: "<XDISPLAY>",
         36: "<ENV>",
         39: "<NewENV>",
         85: "COMPRESS (MCCP)",
         86: "COMPRESS2 (MCCP)",
         91: "MXP"}

# more info on 85/86/MCCP: http://www.randomly.org/projects/MCCP/protocol.html
# more info on 91/MXP: http://www.zuggsoft.com/zmud/mxp.htm

# telnet control codes


# Telnet protocol constants
IAC = "\xFF"  # Interpret As Command
GA = "\xF9"   # Go Ahead
TELOPT_EOR = "\xEF"  # End of Record
#IAC = chr(255)
DONT = chr(254)
DO = chr(253)
WONT = chr(252)
WILL = chr(251)
SB = chr(250)
#GA = chr(249)
NOP = chr(241)
SE = chr(240)
#TELOPT_EOR = chr(239)
SEND = chr(1)
MODE = chr(1)
FORWARDMASK = chr(2)
IS = chr(0)

# some nice strings to help with the telnet control code
# negotiation
DD = DO + DONT
WW = WILL + WONT
DDWW = DD + WW

# telnet option codes
ECHO = chr(1)
SGA = chr(3)
TERMTYPE = chr(24)
EOR = chr(25)
NAWS = chr(31)
LINEMODE = chr(34)
ENV = chr(39)

BELL = chr(7)
#BELL = '\x07'  # Define BELL as a bytes object (Bell character)


def _fcc(code):
    if ord(code) in CODES:
        return CODES[ord(code)]
    return str(code)

def _cc(option):
    """
    Takes in an option string which we peel apart and return a pretty
    string representation of.

    @param option: the option string to convert
    @type  option: string

    @return: the string representation of the code
    @rtype: string
    """
    if len(option) == 3:
        return " ".join([_fcc(m) for m in option])

    return " ".join([_fcc(option[0]), _fcc(option[1]), _fcc(option[2]),
                     option[3:-2], _fcc(option[-2]), _fcc(option[-1])])


class SocketCommunicator:
    """
    The SocketCommunicator handles all incoming and outgoing data from
    and to the mud, telnet control codes, and some data transformations.
    """

    def __init__(self, e, ses, host, port):
        self._engine = e
        self._config = e.getConfigManager()

        self._sessionname = ''
        self._host = host
        self._port = port
        self._sock = None
        self._ansimode = 1
        self._nego_buffer = ''
        self._shutdownflag = 0
        self._session = ses

        self._debug = 0

        # this is the prompt regex that we use to split the incoming text.
        self._prompt_regex = self._buildPromptRegex()

        # this is the regex that we use to split the incoming text.
        delimiters = (IAC + GA, IAC + TELOPT_EOR, "\n")

        # make a regexp matching any of the delimiters above
        self._line_regex = re.compile("(" + "|".join(delimiters) + ")", re.MULTILINE | re.DOTALL)

        # "The server can do delimited prompts" flag
        self._good_prompts = 0

        # handle termtype issues
        if "term" in config.options:
            self._termtype = config.options["term"][0]
        elif "TERM" in os.environ:
            self._termtype = os.environ["TERM"]
        else:
            self._termtype = "lyntin"

        # we keep track of all the telnet stuff we're doing here
        # so we can look at it and dump it or whatever
        self._controllog = []

    def _buildPromptRegex(self, prompt=""):
        """
        Builds the prompt regex.  A prompt is IAC+GA or IAC+TELOPT_EOR or
        any string prompt.  Note that prompts eat up the characters.  So if
        the prompt is ">> " those characters will disappear from the stream.

        Note: the prompt is NOT escaped before it is added to the regexp.  It's
        up to you to re.escape the bits that need escaping.

        @param prompt: the text prompt to use (if any)
        @type  prompt: string

        @returns: the compiled regular expression for prompt detection
        @rtype: Regexp object
        """
        # Escape the prompt if provided
        escaped_prompt = re.escape(prompt) if prompt else ""

        # Build the regex pattern
        if escaped_prompt:
            pattern = f"({IAC}{GA}|{IAC}{TELOPT_EOR}|{escaped_prompt})"
        else:
            pattern = f"({IAC}{GA}|{IAC}{TELOPT_EOR})"

        # Return the compiled regex
        return re.compile(pattern)

    def __repr__(self):
        return "connection %s %d" % (self._host, self._port)

    def logControl(self, str):
        self._controllog.append(str)

    def setSessionName(self, name):
        """
        Sets the session name.

        @param name: the new session name
        @type  name: string
        """
        self._sessionname = name

    def shutdown(self):
        """
        Shuts down the thread polling the socket connection and the socket
        as well.
        """
        self._shutdownflag = 1

    def connect(self, host, port, sessionname):
        """
        Takes in a host and a port and connects the socket.

        @param host: the host to connect to
        @type  host: string

        @param port: the port to connect at
        @type  port: int

        @param sessionname: the name of the new session
        @type  sessionname: string
        """
        if not self._sock:
          sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
          try:
            sock.connect((host, port))  # Attempt to connect to the host and port
            sock.setblocking(True)  # Set the socket to blocking mode
          except socket.error as e:
            print(f"Error connecting to {host}:{port} - {e}")  # Log the error or handle it appropriately
            sock.close()
            raise  # Re-raise the exception after cleanup
          self._sock = sock  # Assign the socket to the instance variable

          self._host = host
          self._port = port
          self._sock = sock
          self._sessionname = sessionname

          ses = exported.get_session(sessionname)

          exported.hook_spam("connect_hook",
                               {"session": ses, "host": host, "port": port})
        else:
            raise Exception("Connection already exists.")

    def _pollForData(self):
        """
        Polls the socket for data. Returns the received data or None if no data is available.
        """
        readers, e, w = select.select([self._sock], [], [], 0.2)

        if readers:
            data = readers[0].recv(1024)
            text = data.decode('latin-1', errors="ignore")
            return text
        else:
            pass

        return None

    def _filterIncomingData(self, data):
        """
    run the data through the net_read_data_filter hook which
    allows things like compressors and other data transformation
    mechanisms to do their thing.
    """
        # be careful--this catches both the '' and the None situations
        if not data:
            return data

        spamargs = {"session": self._session, "data": data, "dataadj": data}
        spamargs = exported.hook_spam("net_read_data_filter",
                                      argmap=spamargs, mappingfunc=exported.filter_mapper)

        if spamargs == None:
            data = ""
        else:
            data = spamargs["dataadj"]

        return data

    def run(self):
        """
    While the connection hasn't been shut down, we spin through this
    loop retrieving data from the mud,
    """
        from lyntin import exported
        try:
            data = ''

            while not self._shutdownflag:
                newdata = self._pollForData()

                if newdata:
                    newdata = self._filterIncomingData(newdata)
                    if newdata == "":
                        continue

                    last_index = 0

                    alldata = (data + newdata).replace("\r", "")
                    # incrementally walk through each line in the data,
                    # adjusting last_index to the end of the previous match
                    for (m) in self._line_regex.finditer(alldata):
                        oneline = alldata[last_index:m.end()]
                        last_index = m.end()
                        self.handleData(oneline)
                    # keep the remainder (empty if alldata ended with a delimiter)
                    data = alldata[last_index:]

                elif newdata == '':
                    # if we got back an empty string, then something's amiss
                    # and we should dump them.
                    if data:
                        self.handleData(data)
                    if self._shutdownflag == 0 and self._session:
                        self._session.shutdown(())
                    break

                elif not self._good_prompts and data:
                    # Now we have rest of the input which is neither delimited prompt
                    # nor complete line, and we yet did not see this server
                    # delimiting it's prompts with telnet GA or EOR option.
                    # We'll handle these data because the socket read was timed out.
                    self.handleData(data)
                    data = ''

        except SystemExit:
            if self._session:
                self._session.shutdown(())

        except:
            exported.write_traceback("socket exception, what broke?")
            if self._session:
                self._session.shutdown(())

        # if we hit this point, we want to shut down the socket
        try:
            self._sock.shutdown(2)
        except:
            pass

        try:
            self._sock.close()
        except:
            pass

        self._sock = None
        self._session = None

        # sometimes the mud will hose up with echo off--we want to kick it
        # on again.
        self._config.change("mudecho", "on")

        # output message so the user knows what happened.
        event.OutputEvent(message.Message(f"Lost connection to: {self._host}\n")).enqueue()

    def write(self, data, convert=True):
        """
    Writes data to the mud after passing it through net_write_data_filter.

    @param data: the data to write to the socket
    @type  data: string

    @param convert: whether (1) or not (0) we should convert eol stuff to
        CRLF and IAC to IAC IAC.
    @type  convert: boolean

    @raises Exception: if we have problems sending the data over the
        socket
    """
        if not self._sock or self._sock.fileno() == -1:
            raise Exception("Socket is not connected or has been closed.")

        if convert:
            data = data.replace("\n", "\r\n")

            if IAC in data:
                data = data.replace(IAC, IAC + IAC)

        if self._shutdownflag == 0:
            # run the data through the net_write_data_filter hook which
            # allows things like compressors and other data transformation
            # mechanisms to do their thing.
            spamargs = {"session": self._session, "data": data, "dataadj": data}
            spamargs = exported.hook_spam("net_write_data_filter", argmap=spamargs,
                                          mappingfunc=exported.filter_mapper)

            if spamargs == None:
                return
            else:
                data = spamargs["dataadj"]

            try:
                data = data.encode("utf-8")
                self._sock.send(data)

            except Exception as e:
                if self._shutdownflag == 0 and self._session:
                    self._session.shutdown(())
                    raise Exception(f"Something went wrong with {e}")

            return None

    def handleData(self, data):
        """
    Handles incoming data from the mud.  We wrap it in a MudEvent
    and toss it on the queue.

    @param data: the incoming data from the mud
    @type  data: string
    """
        global BELL

        # handle the bell
        count = data.count(BELL)
        for i in range(count):
            event.SpamEvent(hookname="bell_hook", argmap={"session": self._session}).enqueue()
        data = data.replace(BELL, "")

        # handle telnet option stuff
        if IAC in data:
            data = self.handleNego(data)

        if not self._config.get("promptdetection") or data.endswith("\n"):
            event.MudEvent(self._session, data).enqueue()
        else:
            event.SpamEvent(hookname="prompt_hook", argmap={"session": self._session, "prompt": data}).enqueue()

    def handleNego(self, data):
        """
    Removes telnet negotiation stuff from the stream and handles it.

    @param data: the incoming data from the mud that we need to parse
        for telnet control code stuff
    @type  data: string

    @return: the data without the telnet control codes
    @rtype:  string
    """
        marker = -1
        i = data.find(IAC)

        while (i != -1):
            if i + 1 >= len(data):
                marker = i
                break

            if data[i + 1] == NOP:
                data = data[:i] + data[i + 2:]
                self.logControl("receive: IAC NOP")

            elif data[i + 1] == GA or data[i + 1] == TELOPT_EOR:
                data = data[:i] + data[i + 2:]
                # if data is a prompt delimited with some telnet option,
                # then we'll mark the server as "server with good prompting"
                self._good_prompts = 1

            elif data[i + 1] == IAC:
                data = data[:i] + data[i + 1:]
                i = i + 1
                self.logControl("receive: IAC IAC")

            else:
                if i + 2 >= len(data):
                    marker = i
                    break

                # handles DO/DONT/WILL/WONT stuff
                if data[i + 1] in DDWW:
                    option = data[i:i + 3]

                    self.logControl("receive: " + _cc(option))
                    if option[2] == ECHO:
                        if option[1] == WILL:
                            self._config.change("mudecho", "off")
                        elif option[1] == WONT:
                            self._config.change("mudecho", "on")

                    elif option[2] == TERMTYPE:
                        if option[1] == DO:
                            self.write(IAC + WILL + TERMTYPE, False)
                            self.logControl("send: IAC WILL TERMTYPE")
                        else:
                            self.write(IAC + WONT + TERMTYPE, False)
                            self.logControl("send: IAC WONT TERMTYPE")

                    elif option[2] == EOR:
                        if option[1] == WILL:
                            self.write(IAC + DO + EOR, False)
                            self.logControl("send: IAC DO EOR")

                    else:
                        args = {"session": self._session, "data": option}
                        # this will give us back the args (in the case that no one
                        # handled it) or None (in the case that someone handled it
                        # and raised a StopSpammingException).
                        ret = exported.hook_spam("net_handle_telnet_option", args)

                        if ret:
                            if option[1] in DD:
                                self.write(IAC + WONT + option[2], False)
                                self.logControl("send: " + _cc(IAC + WONT + option[2]))

                            elif option[1] in WW:
                                self.write(IAC + DONT + option[2], False)
                                self.logControl("send: " + _cc(IAC + DONT + option[2]))

                    data = data[:i] + data[i + 3:]

                # handles SB...SE stuff
                elif data[i + 1] == SB:

                    end = data.find(SE, i)
                    if end == -1:
                        marker = i
                        break

                    option = data[i:end + 1]
                    self.logControl("receive: " + _cc(option))

                    if option[2] == TERMTYPE and option[3] == SEND:
                        self.write(IAC + SB + TERMTYPE + IS + self._termtype + IAC + SE, False)
                        self.logControl("send: IAC SB TERMTYPE IS " + self._termtype + " IAC SE")
                    else:
                        args = {"session": self._session, "data": option}
                        # this will give us back the args (in the case that no one
                        # handled it) or None (in the case that someone handled it
                        # and raised a StopSpammingException).
                        #ret = exported.hook_spam("net_handle_telnet_option", args)
                        exported.hook_spam("net_handle_telnet_option", args)

                    data = data[:i] + data[end + 1:]

                # in case they passed us something weird we remove the IAC and
                # move on
                else:
                    data = data[:i] + data[i + 1:]

            i = data.find(IAC, i)

        if marker != -1:
            self._nego_buffer = data[marker:]
            data = data[:marker]
        return data



# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001-2007
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: __init__.py,v 1.4 2007/07/24 00:39:03 willhelm Exp $
#######################################################################
"""
This is the ui package.  Ui's to be used in Lyntin need to be dropped
in here.  ui's should extend the base.BaseUI class and should also
implement the get_ui_instance function that actually returns an instance 
of the ui in question.

See the textui and tkui code for examples.
"""



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2002-2007
#
# $Id: deed.py,v 1.3 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines the DeedManager which handles deeds (user events).

Deeds can be used to keep track of when things happened.  I tend to
create deeds whenever someone does something nice for me via an action.
This allows me to have a log of who did what to me and when they did it
(if you use the $TIMESTAMP variable) even if I'm away from the computer
or in the midst of copious battle spam.
"""

# deed code originally written by Sebastian John.

import string
from lyntin import manager, utils, exported
from lyntin.modules import modutils

class DeedData:
  def __init__(self):
    self._deeds = []
  
  def addDeed(self, deed):
    """
    Adds a deed to the list.
    
    @param deed: the actual deed text
    @type deed: string
    """
    self._deeds.append(deed)
  
  def clear(self):
    """
    Removes all the deeds.
    """
    self._deeds = []
  
  def getInfo(self, num=""):
    """
    Returns information about the deeds in here.
    
    This is used only by #deed to show all the deeds stored.
    
    @param num: if a number, then returns the last num deeds.  otherwise
        returns deeds that match what this expands to.
    @type num: string or int

    @returns: one big string with all the deeds in it
    @rtype: string
    """
    if not self._deeds:
      return ""
    
    if num.isdigit():
      count = int(num)
      listing = self._deeds[-count:]
    else:
      listing = self._deeds
    
    return string.join(listing, "\n")
  
  def getStatus(self):
    """
    Returns the number of deeds actually stored.

    @returns: the total number of deeds in this DeedData object.
    @rtype: int
    """
    return "%d deed(s)." % len(self._deeds)

class DeedManager(manager.Manager):
  def __init__(self):
    self._deeds = {}

  def getDeedData(self, ses):
    if ses not in self._deeds:
      self._deeds[ses] = DeedData()
    return self._deeds[ses]

  def clear(self, ses):
    if ses in self._deeds:
      self._deeds[ses].clear()

  def getInfo(self, ses, num=""):
    return self.getDeedData(ses).getInfo(num)

  def getStatus(self, ses):
    return self.getDeedData(ses).getStatus()


commands_dict = {}

def deed_cmd(ses, args, input):
  """
  Deeds serve as a kind of notebook - whatever you don't want
  to forget, store it in a deed.

  examples::
    #deed                             -- prints all the deeds for 
                                         that session
    #deed {$TIMESTAMP Joe healed me}  -- adds a new deed to the list
    #deed 10                          -- prints the last 10 deeds

  Before a deed is stored, variables are expanded--this allows you
  to use system, global, and session variables in your deeds like
  $TIMESTAMP which will mark when the deed was created.
  
  category: commands
  """
  # original deed_cmd code contributied by Sebastian John

  if (ses.getName() == "common"):
    exported.write_error("deed cannot be applied to common session.", ses)
    return

  deedtext = args["text"]
  quiet = args["quiet"]

  if not deedtext:
    data = exported.get_manager("deed").getInfo(ses)
    if data == "":
      data = "deed: no deeds defined."
    
    exported.write_message(data, ses)
    return
  
  if deedtext.isdigit():
    data = exported.get_manager("deed").getInfo(ses, deedtext)
    if data == "":
      data = "deed: no deeds defined."
    
    exported.write_message(data, ses)
    return

  exported.get_manager("deed").getDeedData(ses).addDeed(deedtext)
  if not quiet:
    exported.write_message("deed: {%s} added." % deedtext, ses)

commands_dict["deed"] = (deed_cmd, "text= quiet:boolean=false")


dm = None

def load():
  """ Initializes the module by binding all the commands."""
  global dm
  modutils.load_commands(commands_dict)
  dm = DeedManager()
  exported.add_manager("deed", dm)


def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global dm
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("deed")

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:
